# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

class CodedException(Exception):
    def __init__(self, code, error_info=''):
        super().__init__()
        self.code = code
        self.error_info = self.err_strs.get(code) + error_info

    def __str__(self):
        return self.error_info
    
    
class MsprobeException(CodedException):
    INVALID_PARAM_ERROR = 0
    OVERFLOW_NUMS_ERROR = 1
    RECURSION_LIMIT_ERROR = 2
    INTERFACE_USAGE_ERROR = 3
    UNSUPPORTED_TYPE_ERROR = 4

    err_strs = {
        INVALID_PARAM_ERROR: "[msprobe] 无效参数：",
        OVERFLOW_NUMS_ERROR: "[msprobe] 超过预设溢出次数 当前溢出次数：",
        RECURSION_LIMIT_ERROR: "[msprobe] 递归调用超过限制：",
        INTERFACE_USAGE_ERROR: "[msprobe] Invalid interface usage: ",
        UNSUPPORTED_TYPE_ERROR: "[msprobe] Unsupported type: "
    }


class FileCheckException(CodedException):
    INVALID_FILE_ERROR = 0
    FILE_PERMISSION_ERROR = 1
    SOFT_LINK_ERROR = 2
    ILLEGAL_PATH_ERROR = 3
    ILLEGAL_PARAM_ERROR = 4
    FILE_TOO_LARGE_ERROR = 5

    err_strs = {
        SOFT_LINK_ERROR: "[msprobe] 检测到软链接： ",
        FILE_PERMISSION_ERROR: "[msprobe] 文件权限错误： ",
        INVALID_FILE_ERROR: "[msprobe] 无效文件： ",
        ILLEGAL_PATH_ERROR: "[msprobe] 非法文件路径： ",
        ILLEGAL_PARAM_ERROR: "[msprobe] 非法打开方式： ",
        FILE_TOO_LARGE_ERROR: "[msprobe] 文件过大： "
    }


class ParseJsonException(CodedException):
    UnexpectedNameStruct = 0
    InvalidDumpJson = 1
    err_strs = {
        UnexpectedNameStruct: "[msprobe] Unexpected name in json: ",
        InvalidDumpJson: "[msprobe] Invalid dump.json format: ",
    }


class ScopeException(CodedException):
    InvalidApiStr = 0
    InvalidScope = 1
    ArgConflict = 2
    err_strs = {
        InvalidApiStr: "[msprobe] Invalid api_list: ",
        InvalidScope: "[msprobe] Invalid scope: ",
        ArgConflict: "[msprobe] Scope and api_list conflict: ",
    }


class RepairException(CodedException):
    InvalidRepairType = 0
    err_strs = {
        InvalidRepairType: "[msprobe] Invalid repair_type: "
    }


class StepException(CodedException):
    InvalidPostProcess = 0
    err_strs = {
        InvalidPostProcess: "[msprobe] 错误的step后处理配置: ",
    }


class FreeBenchmarkException(CodedException):
    UnsupportedType = 0
    InvalidGrad = 1
    InvalidPerturbedOutput = 2
    OutputIndexError = 3
    err_strs = {
        UnsupportedType: "[msprobe] Free benchmark get unsupported type: ",
        InvalidGrad: "[msprobe] Free benchmark gradient invalid: ",
        InvalidPerturbedOutput: "[msprobe] Free benchmark invalid perturbed output: ",
        OutputIndexError: "[msprobe] Free benchmark output index out of bounds: ",
    }


class DistributedNotInitializedError(Exception):
    def __init__(self, msg):
        super().__init__()
        self.msg = msg

    def __str__(self):
        return self.msg


class ApiAccuracyCheckerException(CodedException):
    ParseJsonFailed = 0
    UnsupportType = 1
    WrongValue = 2
    ApiWrong = 3
    err_strs = {
        ParseJsonFailed: "[msprobe] Api Accuracy Checker parse json failed: ",
        UnsupportType: "[msprobe] Api Accuracy Checker get unsupported type: ",
        WrongValue: "[msprobe] Api Accuracy Checker get wrong value: ",
        ApiWrong: "[msprobe] Api Accuracy Checker something wrong with api: ",
    }



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

def npu_confusion_transpose(data, perm, shape, transpose_first):
    if transpose_first:
        output = data.permute(*perm).contiguous().view(shape)
    else:
        output = data.view(shape).permute(*perm)
    return output


def npu_confusion_transpose_backward(grad, perm, shape, transpose_first):
    try:
        shape_cal = shape if transpose_first else [shape[perm_dim] for perm_dim in perm]
    except IndexError as e:
        raise IndexError("npu_confusion_transpose_backward: Invalid perm index for shape") from e

    perm_cal = [0] * len(perm)
    for i, perm_dim in enumerate(perm):
        perm_cal[perm_dim] = i

    if transpose_first:
        result = grad.permute(*perm_cal).reshape(shape_cal)
    else:
        result = grad.reshape(shape_cal).permute(*perm_cal)
    return result.cpu()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import json
from werkzeug.wrappers.request import Request
from werkzeug import Response
from ..utils.graph_utils import GraphUtils
from ..utils.constant import security_headers


def request_method(allowed_method):
    def decorator(func):
        def wrapper(*args, **kwargs):
            try:
                if len(args) <= 0:
                    raise RuntimeError('Illegal function call, at least 1 parameter is required but got 0')
                request = args[0]
                if not isinstance(request, Request):
                    raise RuntimeError('The request "parameter" is not in a format supported by werkzeug')
                actual_method = request.method.upper()
                expected_method = allowed_method.upper()
                if actual_method != expected_method:
                    raise RuntimeError(f"Method Not Allowed: expected '{expected_method}', got '{actual_method}'")
            except Exception as e:
                result = {'success': False, 'error': str(e)}
                return Response(json.dumps(result), content_type="application/json", headers=security_headers)
            try:
                return func(*args, **kwargs)
            except Exception as e:
                result = {'success': False, 'error': GraphUtils.t('serverError')}
                return Response(json.dumps(result), content_type="application/json", headers=security_headers)
        return wrapper
    return decorator



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves the common function definition.
"""
from enum import Enum


class ShapeType(Enum):
    """
    The enum for shape type
    """
    Scalar = 0
    Vector = 1
    Matrix = 2
    Tensor = 3


class FusionRelation(Enum):
    """
    The enum for fusion relation
    """
    OneToOne = 0
    MultiToOne = 1
    OneToMulti = 2
    MultiToMulti = 3
    L1Fusion = 4


class DatasetAttr(Enum):
    """
    The enum for pytorch dump data attribute
    """
    DataType = 0
    DeviceType = 1
    FormatType = 2
    Type = 3
    Stride = 4


class DeviceType(Enum):
    """
    The enum for device type
    """
    GPU = 1
    NPU = 10
    CPU = 0



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import subprocess
import json
import os
import sys
import argparse
import time
import signal
import threading
from collections import namedtuple
from itertools import cycle
from tqdm import tqdm
from msprobe.pytorch.api_accuracy_checker.acc_check.acc_check import _acc_check_parser, preprocess_forward_content
from msprobe.pytorch.api_accuracy_checker.acc_check.acc_check_utils import get_validated_result_csv_path, \
    get_validated_details_csv_path
from msprobe.pytorch.api_accuracy_checker.compare.compare import Comparator
from msprobe.pytorch.common import parse_json_info_forward_backward
from msprobe.pytorch.common.log import logger
from msprobe.core.common.file_utils import FileChecker, check_file_suffix, check_link, FileOpen, \
    create_directory, load_json, save_json, read_csv
from msprobe.core.common.file_utils import remove_path
from msprobe.core.common.const import FileCheckConst, Const
from msprobe.core.common.utils import CompareException


def split_json_file(input_file, num_splits, filter_api):
    forward_data, backward_data, real_data_path = parse_json_info_forward_backward(input_file)
    input_dir = os.path.dirname(os.path.abspath(input_file))
    if filter_api:
        forward_data = preprocess_forward_content(forward_data)
    for data_name in list(forward_data.keys()):
        forward_data[f"{data_name}.forward"] = forward_data.pop(data_name)
    for data_name in list(backward_data.keys()):
        backward_data[f"{data_name}.backward"] = backward_data.pop(data_name)

    input_data = load_json(input_file)
    if "dump_data_dir" not in input_data.keys():
        logger.error("Invalid input file, 'dump_data_dir' field is missing")
        raise CompareException("Invalid input file, 'dump_data_dir' field is missing")
    if input_data.get("data") is None:
        logger.error("Invalid input file, 'data' field is missing")
        raise CompareException("Invalid input file, 'data' field is missing")
    input_data.pop("data")

    items = list(forward_data.items())
    total_items = len(items)
    chunk_size = total_items // num_splits
    split_files = []

    for i in range(num_splits):
        start = i * chunk_size
        end = (i + 1) * chunk_size if i < num_splits - 1 else total_items

        split_forward_data = dict(items[start:end])
        temp_data = {
            **input_data,
            "data": {
                **split_forward_data,
                **backward_data
            }
        }
        split_filename = os.path.join(input_dir, f"temp_part{i}.json")
        save_json(split_filename, temp_data)
        split_files.append(split_filename)

    return split_files, total_items


def signal_handler(signum, frame):
    logger.warning(f'Signal handler called with signal {signum}')
    raise KeyboardInterrupt()


ParallelUTConfig = namedtuple('ParallelUTConfig', ['api_files', 'out_path', 'num_splits',
                                                   'save_error_data_flag', 'jit_compile_flag', 'device_id',
                                                   'result_csv_path', 'total_items', 'config_path'])


def run_parallel_ut(config):
    processes = []
    device_id_cycle = cycle(config.device_id)
    if config.save_error_data_flag:
        logger.info("acc_check task error data will be saved")
    logger.info(f"Starting parallel acc_check with {config.num_splits} processes")
    progress_bar = tqdm(total=config.total_items, desc="Total items", unit="items")

    def create_cmd(api_info, dev_id):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        acc_check_path = os.path.join(dirname, "acc_check.py")
        cmd = [
            sys.executable, acc_check_path,
            '-api_info', api_info,
            *(['-o', config.out_path] if config.out_path else []),
            '-d', str(dev_id),
            *(['-j'] if config.jit_compile_flag else []),
            *(['-save_error_data'] if config.save_error_data_flag else []),
            '-csv_path', config.result_csv_path,
            *(['-config', config.config_path] if config.config_path else [])
        ]
        return cmd

    def read_process_output(process):
        try:
            while True:
                if process.poll() is not None:
                    break
                output = process.stdout.readline()
                if output == '':
                    break
                if '[ERROR]' in output:
                    logger.warning(output)
                    sys.stdout.flush()
        except ValueError as e:
            logger.warning(f"An error occurred while reading subprocess output: {e}")
        finally:
            if process.poll() is None:
                process.stdout.close()

    def update_progress_bar(progress_bar, result_csv_path):
        while any(process.poll() is None for process in processes):
            result_file = read_csv(result_csv_path)
            completed_items = len(result_file)
            progress_bar.update(completed_items - progress_bar.n)
            time.sleep(1)

    for api_info in config.api_files:
        cmd = create_cmd(api_info, next(device_id_cycle))
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL,
                                   text=True, bufsize=1, shell=False)
        processes.append(process)
        threading.Thread(target=read_process_output, args=(process,), daemon=True).start()

    progress_bar_thread = threading.Thread(target=update_progress_bar, args=(progress_bar, config.result_csv_path))
    progress_bar_thread.start()

    def clean_up():
        progress_bar.close()
        for process in processes:
            try:
                process.terminate()
                process.wait(timeout=1)
            except subprocess.TimeoutExpired:
                process.kill()
        for file in config.api_files:
            check_link(file)
            try:
                remove_path(file)
            except FileNotFoundError:
                logger.warning(f"File not found and could not be deleted: {file}")

    try:
        for process in processes:
            process.communicate(timeout=None)
    except KeyboardInterrupt:
        logger.warning("Interrupted by user, terminating processes and cleaning up...")
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")
    finally:
        # 最后再更新一次进度条，避免因缓存写入等原因子进程结束而进度未刷新的问题
        if wait_for_file_write_complete(config.result_csv_path):
            result_file = read_csv(config.result_csv_path)
            completed_items = len(result_file)
            progress_bar.update(completed_items - progress_bar.n)
        if progress_bar.n < config.total_items:
            logger.warning("The acc_check task has not been completed. The parameter '-csv_path' along with the "
                           "path to " 
                           "the result CSV file will be utilized to resume the acc_check task.")
        clean_up()
        progress_bar_thread.join()
    try:
        comparator = Comparator(config.result_csv_path, config.result_csv_path, False)
        comparator.print_pretest_result()
    except FileNotFoundError as e:
        logger.error(f"Error: {e}")
    except Exception as e:
        logger.error(f"An unexpected error occurred: {e}")


def wait_for_file_write_complete(file_path, timeout=3600):
    last_size = 0
    start_time = time.time()  # 记录开始时间
    while True:
        current_size = os.path.getsize(file_path)
        # 检查是否文件大小未变化
        if current_size == last_size:
            return True  # 文件写入完成，返回 True
        last_size = current_size
        # 检查是否超时
        if time.time() - start_time > timeout:
            logger.error("write the result csv file timeout.")
            return False  # 超时，返回 False
        time.sleep(0.1)  # 适当的延时


def prepare_config(args):
    api_info_file_checker = FileChecker(file_path=args.api_info_file, path_type=FileCheckConst.FILE,
                                        ability=FileCheckConst.READ_ABLE, file_type=FileCheckConst.JSON_SUFFIX)
    api_info = api_info_file_checker.common_check()
    out_path = args.out_path if args.out_path else Const.DEFAULT_PATH
    create_directory(out_path)
    out_path_checker = FileChecker(out_path, FileCheckConst.DIR, ability=FileCheckConst.WRITE_ABLE)
    out_path = out_path_checker.common_check()
    split_files, total_items = split_json_file(api_info, args.num_splits, args.filter_api)
    config_path = args.config_path if args.config_path else None
    if config_path:
        config_path_checker = FileChecker(config_path, FileCheckConst.FILE,
                                          FileCheckConst.READ_ABLE, FileCheckConst.JSON_SUFFIX)
        config_path = config_path_checker.common_check()
    result_csv_path = args.result_csv_path or os.path.join(
        out_path, f"accuracy_checking_result_{time.strftime('%Y%m%d%H%M%S')}.csv")
    if not args.result_csv_path:
        details_csv_path = os.path.join(out_path, f"accuracy_checking_details_{time.strftime('%Y%m%d%H%M%S')}.csv")
        comparator = Comparator(result_csv_path, details_csv_path, False)
    else:
        result_csv_path = get_validated_result_csv_path(args.result_csv_path, 'result')
        details_csv_path = get_validated_details_csv_path(result_csv_path)
    logger.info(f"acc_check task result will be saved in {result_csv_path}")
    logger.info(f"acc_check task details will be saved in {details_csv_path}")
    return ParallelUTConfig(split_files, out_path, args.num_splits, args.save_error_data,
                            args.jit_compile, args.device_id, result_csv_path,
                            total_items, config_path)


def main():
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    parser = argparse.ArgumentParser(description='Run acc_check in parallel')
    _acc_check_parser(parser)
    parser.add_argument('-n', '--num_splits', type=int, choices=range(1, 65), default=8,
                        help='Number of splits for parallel processing. Range: 1-64')
    args = parser.parse_args()
    config = prepare_config(args)
    run_parallel_ut(config)



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from msprobe.core.common.const import Const
from msprobe.core.common.exceptions import ApiAccuracyCheckerException
from msprobe.mindspore.api_accuracy_checker.type_mapping import float_dtype_str_list
from msprobe.mindspore.common.log import logger


def check_and_get_from_json_dict(dict_instance, key, key_description, accepted_type=None, accepted_value=None):
    '''
    Args:
        dict_instance: dict, dict parsed from input json
        key: str
        key_description: str
        accepted_type: tuple
        accepted_value: Union[tuple, list]

    Return:
        value, the corresponding value of "key" in "dict_instance"

    Exception:
        raise ApiAccuracyCheckerException.ParseJsonFailed error when
        1. dict_instance is not a dict
        2. value is None
        3. value is not accepted type
        4. value is not accepted value
    '''
    if not isinstance(dict_instance, dict):
        error_info = "check_and_get_from_json_dict failed: input is not a dict"
        raise ApiAccuracyCheckerException(ApiAccuracyCheckerException.ParseJsonFailed, error_info)
    value = dict_instance.get(key)
    if value is None:
        error_info = f"check_and_get_from_json_dict failed: {key_description} is missing"
        raise ApiAccuracyCheckerException(ApiAccuracyCheckerException.ParseJsonFailed, error_info)
    elif accepted_type is not None and not isinstance(value, accepted_type):
        error_info = f"check_and_get_from_json_dict failed: {key_description} is not accepted type: {accepted_type}"
        raise ApiAccuracyCheckerException(ApiAccuracyCheckerException.ParseJsonFailed, error_info)
    elif accepted_value is not None and value not in accepted_value:
        error_info = f"check_and_get_from_json_dict failed: {key_description} is not accepted value: {accepted_value}"
        raise ApiAccuracyCheckerException(ApiAccuracyCheckerException.ParseJsonFailed, error_info)
    return value


def convert_to_tuple(args):
    if isinstance(args, (tuple, list)):
        return tuple(args)
    else:
        input_list = [args]
        return tuple(input_list)


def trim_output_compute_element_list(compute_element_list, forward_or_backward):
    '''
    Args:
        compute_element_list: List[ComputeElement]
        forward_or_backward: str, Union["forward", "backward"]
    '''
    trimmed_list = []
    for compute_element in compute_element_list:
        if compute_element.get_parameter() is None or \
                (forward_or_backward == Const.BACKWARD and compute_element.get_dtype() not in float_dtype_str_list):
            # trim case: 1. parameter is None. 2. backward output has non float parameter
            continue
        trimmed_list.append(compute_element)
    return trimmed_list


class GlobalContext:
    def __init__(self):
        self.is_constructed = True
        self.dump_data_dir = ""
        self.framework = Const.MS_FRAMEWORK

    def init(self, is_constructed, dump_data_dir, framework):
        self.is_constructed = is_constructed
        self.dump_data_dir = dump_data_dir
        self.framework = framework

    def get_dump_data_dir(self):
        return self.dump_data_dir

    def get_is_constructed(self):
        return self.is_constructed

    def get_framework(self):
        return self.framework


global_context = GlobalContext()



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import time
import json
from tensorboard.util import tb_logging

from .graph_service_base import GraphServiceStrategy
from ..repositories.graph_repo_vis import GraphRepoVis
from ..utils.global_state import GraphState
from ..utils.graph_utils import GraphUtils
from ..model.layout_hierarchy_model import LayoutHierarchyModel
from ..model.match_nodes_model import MatchNodesController
from ..utils.constant import NPU_PREFIX, BENCH_PREFIX, NPU, BENCH, SINGLE, UN_MATCHED_VALUE
from ..utils.constant import MAX_RELATIVE_ERR, MIN_RELATIVE_ERR, MEAN_RELATIVE_ERR, NORM_RELATIVE_ERR

logger = tb_logging.get_logger()


class JsonGraphService(GraphServiceStrategy):

    def __init__(self, run_path, tag):
        super().__init__(run_path, tag)
        self.repo = None

    def load_graph_data(self):
        runs = GraphState.get_global_value('runs')
        run_path = runs.get(self.run)
        buffer = ""
        read_bytes = 0
        chunk_size = 1024 * 1024 * 60  # 缓冲区
        json_data = None  # 最终存储的变量
        _, error = GraphUtils.safe_load_data(run_path, f"{self.tag}.vis", True)
        if error:
            return {'success': False, 'error': GraphUtils.t('loadErrorBySecurityIssue')}
        file_path = os.path.join(run_path, f"{self.tag}.vis")
        file_path = os.path.normpath(file_path)  # 标准化路径
        file_size = os.path.getsize(file_path)  
        if file_size == 0:
            return {'success': False, 'error': '文件为空'}
        with open(file_path, 'r', encoding='utf-8') as f:
            while True:
                chunk = f.read(chunk_size)
                if not chunk:
                    break
                read_bytes += len(chunk)
                buffer += chunk
                current_progress = min(95, int((read_bytes / file_size) * 100)) 
                reading_info = {
                    'progress': current_progress,
                    'status': 'reading',
                    'size': file_size,
                    'read': read_bytes
                }
                yield f"data: {json.dumps(reading_info)}\n\n"
                time.sleep(0.01)  # 避免发送过快

        if json_data is None and buffer:  # 最终验证数据
            try:
                yield f"data: {json.dumps({'progress': current_progress, 'status': 'loading'})}\n\n"
                json_data = GraphUtils.safe_json_loads(buffer)
                yield f"data: {json.dumps({'progress': 99, 'status': 'loading'})}\n\n"
            except json.JSONDecodeError as e:
                err_str = json.dumps({'progress': current_progress, 'error': GraphUtils.t('parseJsonFailed')})
                yield f"data: {err_str}\n\n"

        if json_data is not None:  # 验证存储
            GraphState.set_global_value('current_file_data', json_data)
            GraphState.set_global_value('current_tag', self.tag)
            GraphState.set_global_value('current_run', run_path)
            # 初始化GraphJson
            self.repo = GraphRepoVis(json_data)
            yield f"data: {json.dumps({'done': True, 'progress': 100, 'status': 'loading'})}\n\n"
        else:
            yield f"data: {json.dumps({'progress': current_progress, 'error': GraphUtils.t('parseJsonFailed')})}\n\n"

    def load_graph_config_info(self):
        run_name = self.run
        tag = self.tag
        graph_data, error_message = GraphUtils.get_graph_data({'run': run_name, 'tag': tag})
        if error_message or not graph_data:
            return {'success': False, 'error': str(error_message)}
        config = {}
        try:
            # 读取全局信息,tag层面
            if graph_data.get('MicroSteps', {}):
                config['microSteps'] = graph_data.get('MicroSteps')
            if graph_data.get('ToolTip', {}):
                config['tooltips'] = graph_data.get('ToolTip')
            config['overflowCheck'] = bool(graph_data.get('OverflowCheck')) if 'OverflowCheck' in graph_data else True
            config['isSingleGraph'] = False if graph_data.get(NPU) else True
            # 读取配置信息，run层面
            config_data = GraphState.get_global_value("config_data", {})
            config_data_run = config_data.get(run_name, {})
            if not config_data_run:  # 如果没有run的配置信息，则读取第一个文件中的Colors
                first_run_tags = GraphState.get_global_value("first_run_tags")
                first_tag = first_run_tags.get(run_name)
                if not first_tag:
                    return {'success': False, 'error': GraphUtils.t('loadConfigFailed')}
                first_graph_data, error_message = GraphUtils.get_graph_data({'run': run_name, 'tag': first_tag})
                config_data_run['colors'] = first_graph_data.get('Colors')
                config_data[run_name] = config_data_run
                GraphState.set_global_value('config_data', config_data)
                config['colors'] = first_graph_data.get('Colors')
            else:
                config['colors'] = config_data_run.get('colors')
            # 读取目录下配置文件列表
            config_files = GraphUtils.find_config_files(run_name)
            config['matchedConfigFiles'] = config_files or []
            config['task'] = graph_data.get('task')
            GraphState.set_global_value("config_info", config)
            return {'success': True, 'data': config}
        except Exception as e:
            return {'success': False, 'error': GraphUtils.t('loadConfigFailed')}

    def load_graph_all_node_list(self, meta_data):
        micro_step = meta_data.get('microStep')
        graph_data, error_message = GraphUtils.get_graph_data({'run': self.run, 'tag': self.tag})
        if error_message or not graph_data:
            return {'success': False, 'error': str(error_message)}
        result = {}
        try:
            if not graph_data.get(NPU):
                nodes = GraphUtils.split_graph_data_by_microstep(graph_data, micro_step)
                node_name_list = list(nodes.keys())
                result['npuNodeList'] = node_name_list
                return {'success': True, 'data': result}

            else:
                # 双图
                config_data = GraphState.get_global_value("config_data")
                npu_node = GraphUtils.split_graph_data_by_microstep(graph_data.get(NPU), micro_step)
                bench_node = GraphUtils.split_graph_data_by_microstep(graph_data.get(BENCH), micro_step)
                npu_node_name_list = list(npu_node.keys())
                bench_node_name_list = list(bench_node.keys())
                npu_unmatehed_name_list = [
                    key
                    for key, value in npu_node.items()
                    if not value.get("matched_node_link")
                ]
                bench_unmatehed_name_list = [
                    key
                    for key, value in bench_node.items()
                    if not value.get("matched_node_link")
                ]
                # 保存未匹配和已匹配的节点到全局变量
                config_data['npuUnMatchNodes'] = npu_unmatehed_name_list
                config_data['benchUnMatchNodes'] = bench_unmatehed_name_list
                config_data['npuMatchNodes'] = graph_data.setdefault('npu_match_nodes', {})
                config_data['benchMatchNodes'] = graph_data.setdefault('bench_match_nodes', {})
                # 返回结果
                result['npuNodeList'] = npu_node_name_list
                result['benchNodeList'] = bench_node_name_list
                result['npuUnMatchNodes'] = npu_unmatehed_name_list
                result['benchUnMatchNodes'] = bench_unmatehed_name_list
                result['npuMatchNodes'] = graph_data.setdefault('npu_match_nodes', {})
                result['benchMatchNodes'] = graph_data.setdefault('bench_match_nodes', {})
                for npu_node_name, npu_node in graph_data.get(NPU, {}).get('node', {}).items():
                    if npu_node.get('matched_node_link', None):
                        bench_node_name = npu_node.get('matched_node_link', [None])[-1].replace(BENCH_PREFIX, '', 1)
                        result['npuMatchNodes'][npu_node_name] = bench_node_name
                        result['benchMatchNodes'][bench_node_name] = npu_node_name

                GraphState.set_global_value('config_data', config_data)
                return {'success': True, 'data': result}
        except Exception as e:
            logger.error('get node list error:' + str(e))
            return {'success': False, 'error': GraphUtils.t('getNodeListFailed')}

    def change_node_expand_state(self, node_info, meta_data):

        try:
            graph_data, error_message = GraphUtils.get_graph_data(meta_data)
            if error_message:
                return {'success': False, 'error': str(error_message)}
            if self.repo is None:
                return {'success': False, 'error': GraphUtils.t('initGraphJsonFailed')}
            graph_type = node_info.get('nodeType')
            node_name = node_info.get('nodeName')
            micro_step = meta_data.get('microStep')
            # 单图
            if not graph_data.get(NPU):
                hierarchy = LayoutHierarchyModel.change_expand_state(node_name, SINGLE, self.repo, micro_step, {})
            # NPU
            elif graph_type == NPU:
                hierarchy = LayoutHierarchyModel.change_expand_state(node_name, NPU, self.repo, micro_step, {})
            # 标杆
            elif graph_type == BENCH:
                hierarchy = LayoutHierarchyModel.change_expand_state(node_name, BENCH, self.repo, micro_step, {})
            else:
                return {'success': True, 'data': {}}
            return {'success': True, 'data': hierarchy}
        except Exception as e:
            logger.error('node expand or collapse error:' + str(e))
            node_type_name = ""
            if graph_data.get(NPU):
                node_type_name = GraphUtils.t('debug') if graph_type == NPU else GraphUtils.t('bench')
        return {'success': False, 'error': f"{node_type_name}{GraphUtils.t('expandNodeError')}", 'data': None}

    def search_node_by_precision(self, meta_data, values):
        # 遍历所有的NPU节点，如果节点的精度值在values中，则返回该节点
        graph_data, error_message = GraphUtils.get_graph_data(meta_data)
        if error_message:
            return {'success': False, 'error': str(error_message)}
     
        precision = []
        is_filter_unmatch_nodes = True if UN_MATCHED_VALUE in values else False
        try:
            if is_filter_unmatch_nodes:
                values.remove(UN_MATCHED_VALUE)
            # 单图
            if not graph_data.get(NPU):
                node_list = GraphUtils.split_graph_data_by_microstep(graph_data.get('node', {}),
                                                                     meta_data.get("microStep", -1))
            # 多图
            else:
                node_list = GraphUtils.split_graph_data_by_microstep(graph_data.get(NPU),
                                                                     meta_data.get("microStep", -1))
            for node_name, node in node_list.items():
                subnodes = node.get("subnodes", None)
                if subnodes != [] and subnodes is not None:
                    continue 
                matched_node_link = node.get('matched_node_link', None)
                if is_filter_unmatch_nodes and (matched_node_link is None or matched_node_link == []):
                    precision.append(node_name)
                if any(low <= node.get('data', {}).get("precision_index", -1) < high for low, high in values):
                    precision.append(node_name)
            return {'success': True, 'data': precision}
        except Exception as e:
            logger.error('search precision node failed:' + str(e))
            return {'success': False, 'error': GraphUtils.t('obtainNodesAccuracyFailed')}
    
    def search_node_by_overflow(self, meta_data, values):
        # 遍历所有的NPU节点，如果节点的精度值在values中，则返回该节点
        graph_data, error_message = GraphUtils.get_graph_data(meta_data)
        if error_message:
            return {'success': False, 'error': str(error_message)}
        overflow = []   
        try:
            # 单图
            if not graph_data.get(NPU):
                node_list = GraphUtils.split_graph_data_by_microstep(graph_data.get('node', {}),
                                                                     meta_data.get("microStep", -1))
                for node_name, node in node_list.items():
                    subnodes = node.get("subnodes", None)
                    if subnodes != [] and subnodes is not None:
                        continue 
                    if node.get('data', {}).get("overflow_level", -1) in values:
                        overflow.append(node_name)
                return {'success': True, 'data': overflow}
            else:
                return {'success': False, 'error': GraphUtils.t('notSupportOverflow')}         
        except Exception as e:
            logger.error('search overflow node failed:' + str(e))
            return {'success': False, 'error': GraphUtils.t('obtainNodesOverflowFailed')}

    def update_precision_error(self, meta_data, filter_value):
        try:
            graph_data, error_message = GraphUtils.get_graph_data(meta_data)
            if error_message:
                return {'success': False, 'error': str(error_message)}
            npu_node_list = graph_data.get(NPU, {}).get('node', {})
            for _, node_info in npu_node_list.items():
                output_statistical_diff = node_info.get('output_data', None)
                if not node_info.get('matched_node_link') or not output_statistical_diff:
                    continue
                max_rel_error = -1
                #  根据filter_value 的选择指标计算新的误差值
                for _, diff_values in output_statistical_diff.items():
                    filter_diff_rel = []
                    if MAX_RELATIVE_ERR in filter_value:
                        filter_diff_rel.append(diff_values.get('MaxRelativeErr'))
                    if MIN_RELATIVE_ERR in filter_value:
                        filter_diff_rel.append(diff_values.get('MinRelativeErr'))
                    if NORM_RELATIVE_ERR in filter_value:
                        filter_diff_rel.append(diff_values.get('NormRelativeErr'))
                    if MEAN_RELATIVE_ERR in filter_value:
                        filter_diff_rel.append(diff_values.get('MeanRelativeErr'))
                    # 过滤掉N/A
                    filter_diff_rel = [x for x in filter_diff_rel if x and x != 'N/A']
                    # 如果output指标中存在 Nan/inf/-inf, 直接标记为最大值
                    if "Nan" in filter_diff_rel or "inf" in filter_diff_rel or "-inf" in filter_diff_rel:
                        max_rel_error = 1
                        break
                    filter_diff_rel = [GraphUtils.convert_to_float(x) for x in filter_diff_rel]
                    max_rel_error_for_key = max(filter_diff_rel) if filter_diff_rel else 0
                    max_rel_error = max(max_rel_error, max_rel_error_for_key)
                if max_rel_error != -1:
                    node_info.setdefault('data', {})['precision_index'] = min(max_rel_error, 1)
            return {'success': True, 'data': {}}
        except Exception as e:
            logger.error('update precision error error:' + str(e))
            return {'success': False, 'error': GraphUtils.t('updatePrecisionFailed')}

    def update_hierarchy_data(self, graph_type):
        if (graph_type == NPU or graph_type == BENCH):
            hierarchy = LayoutHierarchyModel.update_hierarchy_data(graph_type)
            return {'success': True, 'data': hierarchy}
        else:
            return {'success': False, 'error': GraphUtils.t('nodeTypeError')}
 
    def get_node_info(self, node_info, meta_data):
        graph_data, error_message = GraphUtils.get_graph_data(meta_data)
        if error_message:
            return {'success': False, 'error': str(error_message)}
        try:
            graph_type = node_info.get('nodeType')
            node_name = node_info.get('nodeName')
            result = {}
            if not graph_data.get(NPU) or graph_type == SINGLE:
                result['npu'] = graph_data.get('node', {}).get(node_name)
            else:
                matched_node_type = BENCH if graph_type == NPU else NPU
                matched_node_preifx = NPU_PREFIX if matched_node_type == NPU else BENCH_PREFIX
                node = graph_data.get(graph_type, {}).get('node', {}).get(node_name, {})
                matched_node_link = node.get('matched_node_link', []) if isinstance(node.get('matched_node_link', []),
                                                                                    list) else []
                matched_node_name = matched_node_link[-1].replace(matched_node_preifx, '',
                                                                  1) if matched_node_link else None
                matched_node = graph_data.get(matched_node_type, {}).get('node', {}).get(
                    matched_node_name) if matched_node_name else None
                result['npu'] = node if graph_type == NPU else matched_node
                result['bench'] = node if graph_type == BENCH else matched_node
            return {'success': True, 'data': result}
        except Exception as e:
            logger.error('get node info error:' + str(e))
            return {'success': False, 'error': GraphUtils.t('getNodeInfoError'), 'data': None}

    def add_match_nodes(self, npu_node_name, bench_node_name, meta_data, is_match_children):
        graph_data, error_message = GraphUtils.get_graph_data(meta_data)
        if error_message:
            return {'success': False, 'error': str(error_message)}
        task = graph_data.get('task')
        result = {}
        try:
            # 根据任务类型计算误差
            if task == 'md5' or task == 'summary':
                if is_match_children:
                    match_result = MatchNodesController.process_task_add_child_layer(graph_data,
                                                                    npu_node_name, bench_node_name, task)              
                else:
                    match_result = MatchNodesController.process_task_add(graph_data,
                                                                         npu_node_name, bench_node_name, task)
                return self._generate_matched_result(match_result)
            else:
                return {'success': False, 'error': GraphUtils.t('taskTypeError')}
        except Exception as e:
            return {'success': False, 'error': GraphUtils.t('operationFailed'), 'data': None}

    def add_match_nodes_by_config(self, config_file_name, meta_data):
        graph_data, error_message = GraphUtils.get_graph_data(meta_data)
        if error_message:
            return {'success': False, 'error': GraphUtils.t('readFileFailed')}
        match_node_links, error = GraphUtils.safe_load_data(meta_data.get('run'), config_file_name)
        if error:
            return {'success': False, 'error': GraphUtils.t('loadConfigFileError')}
        if not match_node_links:
            return {'success': False, 'error': GraphUtils.t('matchNodeLinksNullError')}
        task = graph_data.get('task')
        try:
            # 根据任务类型计算误差
            if task == 'md5' or task == 'summary':
                match_result = MatchNodesController.process_task_add_child_layer_by_config(graph_data,
                                                                                           match_node_links, task)
                result = self._generate_matched_result(match_result)
                if result.get('data'):
                    result['data']['matchResult'] = [item.get('success', False) for item in match_result]
                return result
            else:
                return {'success': False, 'error': GraphUtils.t('taskTypeError')}
        except Exception as e:
            logger.error(str(e))
            return {'success': False, 'error': GraphUtils.t('operationFailed'), 'data': None}

    def delete_match_nodes(self, npu_node_name, bench_node_name, meta_data, is_unmatch_children):
        graph_data, error_message = GraphUtils.get_graph_data(meta_data)
        if error_message:
            return {'success': False, 'error': str(error_message)}
        task = graph_data.get('task')
        try:
            # 根据任务类型计算误差
            if task == 'md5' or task == 'summary':
                if is_unmatch_children:
                    match_result = MatchNodesController.process_task_delete_child_layer(graph_data, npu_node_name,
                                                                                  bench_node_name, task)
                else:
                    match_result = MatchNodesController.process_task_delete(graph_data, npu_node_name,
                                                                            bench_node_name, task)
                
                return self._generate_matched_result(match_result)
            else:
                return {'success': False, 'error': GraphUtils.t('taskTypeError')}
        except Exception as e:
            return {'success': False, 'error': GraphUtils.t('operationFailed'), 'data': None}

    def save_data(self, meta_data):
        if not meta_data:
            return {'success': False, 'error': GraphUtils.t('parameterEmpty')}
        graph_data, error_message = GraphUtils.get_graph_data(meta_data)
        if error_message:
            return {'success': False, 'error': str(error_message)}

        try:
            _, error = GraphUtils.safe_save_data(graph_data, self.run, f"{self.tag}.vis")
            if error:
                return {'success': False, 'error': error}
        except (ValueError, IOError, PermissionError) as e:
            return {'success': False, 'error': GraphUtils.t('saveFileFailed')}
        return {'success': True}

    def update_colors(self, colors):
        """Set new colors in jsondata."""
        try:
            config_data = GraphState.get_global_value("config_data", {})
            first_run_tags = GraphState.get_global_value("first_run_tags")
            config_data_run = config_data.get(self.run, {})
            first_run_tag = first_run_tags.get(self.run)
            first_file_data, error = GraphUtils.safe_load_data(self.run, f"{first_run_tag}.vis")
            if error:
                logger.error(f"Error loading data: {error}")
                return {'success': False, 'error': GraphUtils.t('loadConfigFailed')}
            first_file_data['Colors'] = colors
            config_data_run['colors'] = colors
            config_data[self.run] = config_data_run
            GraphState.set_global_value("config_data", config_data)
            GraphUtils.safe_save_data(first_file_data, self.run, f"{first_run_tag}.vis")
            return {'success': True, 'error': None, 'data': {}}
        except Exception as e:
            return {'success': False, 'error': GraphUtils.t('updateColorFailed'), 'data': None}

    def save_matched_relations(self, meta_data):
        run = meta_data.get('run')
        tag = meta_data.get('tag')
        config_data = GraphState.get_global_value("config_data")
        # 匹配列表和未匹配列表
        npu_match_nodes_list = config_data.get('manualMatchNodes', {})
        try:
            _, error = GraphUtils.safe_save_data(npu_match_nodes_list, run, f"{tag}.vis.config")
            if error:
                return {'success': False, 'error': error}
            else:
                return {'success': True, 'data': f"{tag}.vis.config"}
        except (ValueError, IOError, PermissionError) as e:
            return {'success': False, 'error': GraphUtils.t('operationFailed')}
  
    def _generate_matched_result(self, match_result):
        update_data = []
        for item in match_result:
            if item.get('success') is True:
                for node in item.get('data', []):
                    update_data.append(node)
            
        if len(update_data) > 0:
            config_data = GraphState.get_global_value("config_data")
            result = {
                'success': True,
                'data': {
                    'npuMatchNodes': config_data.get('npuMatchNodes', {}),
                    'benchMatchNodes': config_data.get('benchMatchNodes', {}),
                    'npuUnMatchNodes': config_data.get('npuUnMatchNodes', []),
                    'benchUnMatchNodes': config_data.get('benchUnMatchNodes', [])
                }
            }     
        else:
            result = {'success': False, 'error': GraphUtils.t('selectedNodeCantMatched')}
        return result



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import torch


def npu_rms_norm(x, gamma, epsilon=1e-5):
    rstd = torch.rsqrt(torch.mean(torch.pow(x, 2), axis=-1, keepdim=True) + epsilon)
    res = x * rstd * gamma
    return res, rstd.float()


def npu_rms_norm_backward(grad, x, gamma, rstd):
    mean_gy = (grad * x * gamma * rstd).mean(dim=-1, keepdim=True)
    grad_x = (grad * gamma - x * rstd * mean_gy) * rstd
    grad_gamma = x * grad * rstd
    return grad_x.cpu(), grad_gamma.cpu()




# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This class mainly involves the main function.
"""

import csv
import os
import stat
import time

import pandas as pd

from msprobe.core.common.log import logger
from msprobe.core.common.file_utils import create_directory
from msprobe.infer.offline.compare.msquickcmp.adapter_cli.args_adapter import CmpArgsAdapter
from msprobe.infer.offline.compare.msquickcmp.atc import atc_utils
from msprobe.infer.offline.compare.msquickcmp.common import utils
from msprobe.infer.offline.compare.msquickcmp.common.convert import convert_npy_to_bin
from msprobe.infer.offline.compare.msquickcmp.common.utils import AccuracyCompareException, \
    get_shape_to_directory_name, OPTYPE_WHITWLIST
from msprobe.infer.offline.compare.msquickcmp.net_compare.net_compare import NetCompare
from msprobe.infer.offline.compare.msquickcmp.npu.npu_dump_data import NpuDumpData
from msprobe.infer.offline.compare.msquickcmp.npu.om_parser import OmParser
from msprobe.infer.utils.file_open_check import ms_open, sanitize_csv_value
from msprobe.infer.utils.check.rule import Rule
from msprobe.infer.utils.constants import TENSOR_MAX_SIZE

try:
    import acl
except ImportError:
    acl = None
    logger.error("Please verify that the CANN environment is properly configured.")

WRITE_MODES = stat.S_IWUSR | stat.S_IRUSR
READ_WRITE_FLAGS = os.O_RDWR | os.O_CREAT
WRITE_FLAGS = os.O_WRONLY | os.O_CREAT | os.O_TRUNC
MAX_MEMORY_USE = 6 * 1024 * 1024 * 1024
COSINE_SIMILARITY = 0.99
RELATIVE_EUCLIDEAN_DISTANCE = 0.05
KULLBACK_LEIBLER_DIVERGENCE = 0.005
ROOT_MEAN_SQUARE_ERROR = 1.0
MEAN_RELATIVE_ERROR = 1.0
YES = "YES"
NO = "NO"


def _generate_golden_data_model(args, npu_dump_npy_path):
    model_name, extension = utils.get_model_name_and_extension(args.golden_path)
    if ".onnx" == extension:
        from msprobe.infer.offline.compare.msquickcmp.onnx_model.onnx_dump_data import OnnxDumpData
        return OnnxDumpData(args, npu_dump_npy_path), extension
    elif ".om" == extension:
        return NpuDumpData(arguments=args, is_golden=True), extension
    else:
        logger.error("Only model files whose names end with .om or .onnx are supported")
        raise AccuracyCompareException(utils.ACCURACY_COMPARISON_MODEL_TYPE_ERROR)


def _correct_the_wrong_order(left_index, right_index, golden_net_output_info):
    if left_index not in golden_net_output_info.keys() or right_index not in golden_net_output_info.keys():
        return
    if left_index != right_index:
        tmp = golden_net_output_info[left_index]
        golden_net_output_info[left_index] = golden_net_output_info[right_index]
        golden_net_output_info[right_index] = tmp
        logger.info(f"swap the {left_index} and {right_index} item in golden_net_output_info!")


def _check_output_node_name_mapping(original_net_output_node, golden_net_output_info):
    for left_index, node_name in original_net_output_node.items():
        match = False
        for right_index, dump_file_path in golden_net_output_info.items():
            dump_file_name = os.path.basename(dump_file_path)
            if dump_file_name.startswith(node_name.replace("/", "_").replace(":", ".")):
                match = True
                _correct_the_wrong_order(left_index, right_index, golden_net_output_info)
                break
        if not match:
            logger.warning(f"the original name: {node_name} of net output maybe not correct!")
            break


def _get_single_csv_in_folder(csv_path):
    for file_name in os.listdir(csv_path):
        if file_name.endswith('.csv'):
            return os.path.join(csv_path, file_name)
    raise IOError(f"None csv file exists in folder {csv_path}")


def _read_and_process_csv(csv_path, process_func, node_output_show_list):
    if Rule.input_file().check(csv_path):
        with ms_open(csv_path, 'r', max_size=TENSOR_MAX_SIZE) as f:
            reader = csv.reader(f)
            rows = [row for row in reader]
        if len(rows) < 1:
            logger.error("csv is empty, please check.")
            raise AccuracyCompareException(utils.ACCURACY_COMPARISON_EMPTY_CSV_ERROR)
        header = rows[0]
        rows = process_func(header, rows, node_output_show_list)
    return rows


def _write_csv(csv_path, rows):
    with ms_open(csv_path, mode='w') as f:
        writer = csv.writer(f)
        for line in rows:
            for ele in line:
                _ = sanitize_csv_value(ele)
        writer.writerows(rows)


def _process_is_npu_and_is_precision_error_ops(header, rows, node_output_name_list):
    ground_truth_col = header.index("GroundTruth")
    optype_col = header.index("OpType")
    cosine_similarity_col = header.index("CosineSimilarity")
    relative_euclidean_distance_col = header.index("RelativeEuclideanDistance")
    kullback_leibler_divergence_col = header.index("KullbackLeiblerDivergence")
    root_mean_square_error_col = header.index("RootMeanSquareError")
    mean_relative_error_col = header.index("MeanRelativeError")

    header.append('IsNpuOps')
    header.append('IsOutputNode')
    header.append('IsPrecisionError')
    for row in rows[1:]:
        try:
            is_npu_ops = YES if row[ground_truth_col] == "*" else NO
            row.append(is_npu_ops)

            optype = row[optype_col]
            cosine_similarity = row[cosine_similarity_col]
            relative_euclidean_distance = float(row[relative_euclidean_distance_col])
            kullback_leibler_divergence = float(row[kullback_leibler_divergence_col])
            root_mean_square_error = float(row[root_mean_square_error_col])
            mean_relative_error = float(row[mean_relative_error_col])
            if _is_output_node(row[ground_truth_col], node_output_name_list):
                row.append(YES)
            else:
                row.append(NO)
            if optype in OPTYPE_WHITWLIST or cosine_similarity.lower() == 'nan':
                row.append(NO)
                continue
            if _is_row_precision_error(cosine_similarity,
                                       relative_euclidean_distance,
                                       kullback_leibler_divergence,
                                       root_mean_square_error,
                                       mean_relative_error):
                row.append(YES)
            else:
                row.append(NO)

        except ValueError as e:
            logger.warning(f"Skipping row due to invalid data: {row}. Error: {e}")
            continue
    return rows


def _is_row_precision_error(cosine_similarity,
                            relative_euclidean_distance,
                            kullback_leibler_divergence,
                            root_mean_square_error,
                            mean_relative_error):
    return (float(cosine_similarity) < COSINE_SIMILARITY or
            relative_euclidean_distance > RELATIVE_EUCLIDEAN_DISTANCE or
            kullback_leibler_divergence > KULLBACK_LEIBLER_DIVERGENCE or
            root_mean_square_error > ROOT_MEAN_SQUARE_ERROR or
            mean_relative_error > MEAN_RELATIVE_ERROR)


def _is_output_node(groundtruth, onnxnode_output_name_list):
    return groundtruth.strip() in onnxnode_output_name_list


def _append_column_to_csv(csv_path, node_output_show_list=None):
    if node_output_show_list is None:
        node_output_show_list = []
    csv_path = _get_single_csv_in_folder(csv_path)
    rows = _read_and_process_csv(csv_path, _process_is_npu_and_is_precision_error_ops, node_output_show_list)
    _write_csv(csv_path, rows)


def cmp_process(args: CmpArgsAdapter):
    args.golden_path = os.path.realpath(args.golden_path)
    args.target_path = os.path.realpath(args.target_path)
    args.cann_path = os.path.realpath(args.cann_path)
    args.input_data = convert_npy_to_bin(args.input_data)
    try:
        check_and_run(args)
    except utils.AccuracyCompareException as error:
        raise error


def run(args: CmpArgsAdapter, input_shape, original_out_path):
    if input_shape:
        args.input_shape = input_shape
        args.output_path = os.path.join(original_out_path, get_shape_to_directory_name(args.input_shape))
    run_om_model_compare(args)


def run_om_model_compare(args):
    # whether use aipp
    output_json_path = atc_utils.convert_model_to_json(args.cann_path, args.target_path, args.output_path)
    golden_json_path = None
    if args.golden_path.endswith('.om'):
        golden_json_path = atc_utils.convert_model_to_json(args.cann_path, args.golden_path, args.output_path)

    temp_om_parser = OmParser(output_json_path)
    use_aipp = True if temp_om_parser.get_aipp_config_content() else False

    npu_dump = NpuDumpData(args, is_golden=False)
    # generate npu inputs data
    npu_dump.generate_inputs_data(use_aipp=use_aipp)
    # generate npu dump data
    npu_dump_data_path, npu_net_output_data_path = npu_dump.generate_dump_data()
    npu_dump_npy_path = ""

    # generate onnx inputs data
    golden_dump, model_extension = _generate_golden_data_model(args, npu_dump_npy_path)
    expect_net_output_node = npu_dump.get_expect_output_name()
    # generate dump data by golden model
    golden_dump.generate_inputs_data(npu_dump_data_path, use_aipp)
    if isinstance(golden_dump, NpuDumpData):
        golden_dump_data_path, _ = golden_dump.generate_dump_data(npu_dump_npy_path, npu_dump.om_parser)
    else:
        golden_dump_data_path = golden_dump.generate_dump_data(npu_dump_npy_path, npu_dump.om_parser)
    golden_net_output_info = golden_dump.get_net_output_info()

    # if it's dynamic batch scenario, golden data files should be renamed
    utils.handle_ground_truth_files(npu_dump.om_parser, npu_dump_data_path, golden_dump_data_path)

    # compare the entire network
    net_compare = NetCompare(npu_dump_data_path, golden_dump_data_path, output_json_path, args, golden_json_path)
    net_compare.accuracy_network_compare()

    # Check and correct the mapping of net output node name.
    if len(expect_net_output_node) == 1:
        _check_output_node_name_mapping(expect_net_output_node, golden_net_output_info)

    node_output_show_list = None
    if model_extension == ".onnx":
        node_output_show_list = _get_model_output_node_name_list(golden_dump.model_with_inputs_session,
                                                                 golden_dump.origin_model)
    _append_column_to_csv(args.output_path, node_output_show_list)


def _get_model_output_node_name_list(model_with_inputs_session, origin_model):
    net_output_node_name_list = [item.name for item in model_with_inputs_session.get_outputs()]
    node_output_show_list = []
    for node_name in net_output_node_name_list:
        pre_node = _find_previous_node(origin_model.graph, node_name)
        if pre_node is None:
            return None
        node_output_show_list.append(pre_node)
    return node_output_show_list


def _find_previous_node(graph, output_name):
    # 遍历所有节点
    for node in graph.node:
        if output_name in [output for output in node.output]:
            # 找到目标输出节点
            return node.name
    return None


def check_and_run(args: CmpArgsAdapter):
    time_dir = time.strftime("%Y%m%d%H%M%S", time.localtime())
    original_out_path = os.path.realpath(os.path.join(args.output_path, time_dir))
    args.output_path = original_out_path
    create_directory(args.output_path)

    # deal with the dymShape_range param if exists
    input_shapes = []
    if args.dym_shape_range:
        input_shapes = utils.parse_dym_shape_range(args.dym_shape_range)
    if not input_shapes:
        input_shapes.append("")
    for input_shape in input_shapes:
        run(args, input_shape, original_out_path)
    if args.dym_shape_range:
        csv_sum(original_out_path)


def csv_sum(original_out_path):
    """
    Function:
        Summarize csv files under different shapes
        generate a xlsx file
    """
    csv_file_list = []
    sheet_name_list = []

    for files in os.listdir(original_out_path):
        if not os.path.isdir(os.path.join(original_out_path, files)):
            continue
        for sub_file in os.listdir(os.path.join(original_out_path, files)):
            if sub_file.endswith(".csv"):
                csv_file_list.append(os.path.join(original_out_path, files, sub_file))
                sheet_name_list.append(files)

    xlsx_file_summary = os.path.join(original_out_path, "result_summary.xlsx")

    if os.path.exists(xlsx_file_summary):
        logger.error(f"Error, file {xlsx_file_summary} already exists!")
        os.remove(xlsx_file_summary)

    with ms_open(xlsx_file_summary, 'wb') as fp_write:
        with pd.ExcelWriter(xlsx_file_summary) as writer:
            for i, csv_file in enumerate(csv_file_list):
                if Rule.input_file().check(csv_file):
                    data = pd.read_csv(csv_file, na_values=['NAN'])
                    data.to_excel(writer, sheet_name=sheet_name_list[i], index=False, na_rep='NAN')



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import functools
import threading
from contextlib import nullcontext

import torch

from msprobe.core.common.const import Const
from msprobe.core.common.log import logger
from msprobe.core.common.runtime import Runtime
from msprobe.core.common.utils import replace_last_occurrence, ThreadSafe
from msprobe.core.dump.data_dump.data_processor.base import ModuleForwardInputsOutputs
from msprobe.core.dump.hook_manager import BaseHookManager, HookSet
from msprobe.pytorch.common.utils import (
    torch_version_above_or_equal_2,
    register_forward_hook,
    Const as PtConst
)
from msprobe.pytorch.dump.api_dump.hook_module import HOOKModule
from msprobe.pytorch.dump.module_dump.module_processor import ModuleProcessor

class PytorchHookManager(BaseHookManager):
    @staticmethod
    def _no_grad_context():
        return nullcontext()

    @staticmethod
    def _add_count(name):
        HOOKModule.add_module_count(name)

    @staticmethod
    def _get_count(name):
        return HOOKModule.get_module_count(name)

    @staticmethod
    def _process_kwargs_and_output(module, tid, hook_type, kwargs_or_output, output_or_kwargs):
        if hook_type == Const.API:
            kwargs = kwargs_or_output
            output = output_or_kwargs
        else:
            kwargs = kwargs_or_output if torch_version_above_or_equal_2 else {}
            output = output_or_kwargs if torch_version_above_or_equal_2 else kwargs_or_output
        return kwargs, output

    def build_hook(self, hook_type, name):
        if hook_type == Const.API:
            hook_set = HookSet(
                forward_pre_hook=self._build_forward_pre_hook(hook_type, name),
                distributed_forward_hook=self._build_distributed_forward_hook()
            )
        else:
            full_backward_name = replace_last_occurrence(name, Const.FORWARD, Const.BACKWARD)
            hook_set = HookSet(
                forward_hook=self._build_forward_hook(hook_type, name),
                backward_hook=self._build_backward_hook(hook_type, full_backward_name)
            )
        return hook_set

    def _register_forward_hook(self, module, api_name):
        if not hasattr(module, 'msprobe_forward_hook'):
            register_forward_hook(module, self._build_forward_hook(Const.API, api_name))
            setattr(module, 'msprobe_forward_hook', True)

    def _register_backward_hook(self, module, full_backward_name, args):
        pass

    def _register_backward_pre_hook(self, module, full_backward_name, args, kwargs, output):
        if module.prefix_api_name in PtConst.DROPOUT_API_LIST:
            p = kwargs.get('p', 0.5)
            if len(args) >= 2:
                p = args[1]
            if p == 0:
                logger.debug(f"The parameter 'p' is set to 0, so the {full_backward_name} data will not be collected.")
                return output

        var = output
        while not isinstance(var, torch.Tensor):
            if isinstance(var, dict):
                var = next((v for v in var.values() if isinstance(v, torch.Tensor)))
            elif isinstance(var, (list, tuple)):
                if var:
                    var = var[0]
                else:
                    return output
            else:
                return output

        if not (var.requires_grad and torch.is_grad_enabled()):
            return output

        grad_fn = var.grad_fn
        if grad_fn is not None:
            backward_hook = self._build_backward_hook(Const.API, full_backward_name)
            wrapper = functools.partial(backward_hook, module)
            functools.update_wrapper(wrapper, backward_hook)
            grad_fn.register_hook(wrapper)

        return output

    def _need_exchange(self, module):
        return True

    def _get_params_dict(self, module):
        params_dict = {}
        if self.config.task != Const.STRUCTURE:
            params_dict = {
                key.split(Const.SEP)[-1]: value
                for key, value in module.named_parameters(recurse=False)
            }
        return params_dict

    def _build_distributed_forward_hook(self):
        def distributed_forward_hook(module, full_name, args, kwargs, output):
            if not full_name or not Runtime.is_running:
                return

            tid = threading.get_ident()
            with ThreadSafe():
                BaseHookManager.inner_switch[tid] = True
                self.data_collector.update_api_or_module_name(full_name)
                module_input_output = ModuleForwardInputsOutputs(args=args, kwargs=kwargs, output=output)
                with self._no_grad_context():
                    self.data_collector.forward_output_data_collect(
                        full_name,
                        module,
                        self._pid,
                        module_input_output
                    )
                BaseHookManager.inner_switch[tid] = False

        return distributed_forward_hook

    def _register_param_hook(self, name, module, params_dict):
        ori_name = name.rsplit(Const.SEP, 2)[0]
        # data_mode为forward时，不注册参数hook
        if not (Const.FORWARD in self.config.data_mode and Const.BACKWARD not in self.config.data_mode):
            for param_name, param in params_dict.items():
                if param.requires_grad:
                    name = ori_name + Const.SEP + param_name
                    old_handle = BaseHookManager.hook_handle_dict.get(name)
                    if old_handle and hasattr(old_handle, "remove"):
                        old_handle.remove()
                    param_tmp = param.expand_as(param)
                    if param_tmp.grad_fn:
                        grad_acc = param_tmp.grad_fn.next_functions[0][0]
                        handle = grad_acc.register_hook(self._build_grad_hook(ori_name, param_name, param))
                        BaseHookManager.hook_handle_dict[name] = handle

    def _build_grad_hook(self, ori_name, param_name, param):
        def hook_fn(*args):
            tid = threading.get_ident()
            if not self._should_execute_hook(Const.MODULE, tid):
                return
            with ThreadSafe():
                original_state = self.ensure_gc_enabled()
                BaseHookManager.inner_switch[tid] = True
                index = self._get_grad_hook_call_index(ori_name, param_name)
                if ModuleProcessor.is_megatron_module and hasattr(param, "main_grad"):
                    grad = param.main_grad
                else:
                    grad = param.grad
                self.data_collector.params_data_collect(ori_name, param_name, self._pid, grad, str(index))
                BaseHookManager.inner_switch[tid] = False
                self.restore_gc_state(original_state)
            return

        return hook_fn


# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
from abc import ABC, abstractmethod

from tensorboard.util import tb_logging
from ..utils.graph_utils import GraphUtils
from ..utils.global_state import GraphState
from ..utils.constant import Extension, DataType

logger = tb_logging.get_logger()
DB_EXT = Extension.DB.value
JSON_EXT = Extension.JSON.value
DB_TYPE = DataType.DB.value
JSON_TYPE = DataType.JSON.value


class GraphServiceStrategy(ABC):
    run = ''
    tag = ''

    def __init__(self, run, tag):
        self.run = run
        self.tag = tag

    @staticmethod
    def load_meta_dir():
        """
        Scan logdir for directories containing .vis(.db) files. If the directory contains .vis.db files,
        it is considered a db type and the .vis files are ignored. Otherwise, it is considered a json type.
        """
        logdir = GraphState.get_global_value('logdir')
        runs = GraphState.get_global_value('runs', {})
        first_run_tags = GraphState.get_global_value('first_run_tags', {})
        meta_dir = {}
        error_list = []
        success, error = GraphUtils.safe_check_load_file_path(logdir, True)
        if not success:
            error_list.append({
                'run': logdir,
                'tag': '',
                'info': f'Error logdir:  {str(error)}'
            })
            result = {
                'data': meta_dir,
                'error': error_list
            }
            return {'success': False, 'error': error_list}
        for root, _, files in GraphUtils.walk_with_max_depth(logdir, 2):
            run_abs = os.path.abspath(root)
            run = os.path.basename(run_abs)  # 不允许同名目录，否则有问题
            for file in files:
                if file.endswith(DB_EXT):
                    tag = file[:-len(DB_EXT)]
                    _, error = GraphUtils.safe_load_data(run_abs, file, True)
                    if error:
                        error_list.append({
                            'run': run,
                            'tag': tag,
                            'info': f'Error: {str(error)}'
                        })
                        logger.error(f'Error: File run:"{run_abs},tag:{tag}" is not accessible. Error: {error}')
                        continue
                    runs[run] = run_abs
                    meta_dir[run] = {'type': DB_TYPE, 'tags': [tag]}
                    break
                # 只取文件名，不包含扩展名
                if file.endswith(JSON_EXT):
                    tag = os.path.splitext(file)[0]
                    _, error = GraphUtils.safe_load_data(run_abs, file, True)
                    if error:
                        error_list.append({
                            'run': run,
                            'tag': tag,
                            'info': f'Error: {error}'
                        })
                        logger.error(f'Error: File run:"{run_abs},tag:{tag}" is not accessible. Error: {error}')
                        continue
                    runs[run] = run_abs
                    if meta_dir.get(run) is None:
                        meta_dir[run] = {'type': JSON_TYPE, 'tags': [tag]}
                    else:
                        meta_dir.get(run).get('tags').append(tag)
        meta_dir = GraphUtils.sort_data(meta_dir)
        for run, value in meta_dir.items():
            first_run_tags[run] = value.get('tags')[0] if value.get('tags') else ''
        GraphState.set_global_value('runs', runs)
        GraphState.set_global_value('first_run_tags', first_run_tags)
        result = {
            'data': meta_dir,
            'error': error_list
        }
        return result

    @abstractmethod
    def load_graph_data(self):
        pass

    @abstractmethod
    def load_graph_config_info(self):
        pass

    @abstractmethod
    def load_graph_all_node_list(self, meta_data):
        pass

    @abstractmethod
    def change_node_expand_state(self, node_info, meta_data):
        pass
        
    def search_node_by_precision(self, meta_data, values):
        pass
    
    def search_node_by_overflow(self, meta_data, values):
        pass
    
    @abstractmethod
    def get_node_info(self, node_info, meta_data):
        pass

    @abstractmethod
    def add_match_nodes(self, npu_node_name, bench_node_name, meta_data, is_match_children):
        pass

    @abstractmethod
    def add_match_nodes_by_config(self, config_file_name, meta_data):
        pass

    @abstractmethod
    def delete_match_nodes(self, npu_node_name, bench_node_name, meta_data, is_unmatch_children):
        pass
    
    @abstractmethod
    def update_precision_error(self, meta_data, filter_value):
        pass    

    @abstractmethod
    def update_colors(self, colors):
        pass

    @abstractmethod
    def save_matched_relations(self, meta_data):
        pass
    



# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import os
import shutil
import subprocess
import sys
import tempfile
import glob
import zipfile
import logging

# 配置日志
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s'
)
logger = logging.getLogger(__name__)


class SecureBuildManager:
    """安全构建管理器"""
    
    def __init__(self, source_dir):
        self.source_dir = source_dir
        self.original_cwd = os.getcwd()
    
    @staticmethod
    def _fix_whl_permissions(whl_file_path):
        """修复whl文件内权限"""
        temp_file_path = whl_file_path + ".new"
        
        try:
            with zipfile.ZipFile(whl_file_path, 'r') as source_zip:
                with zipfile.ZipFile(temp_file_path, 'w') as target_zip:
                    for zip_item in source_zip.infolist():
                        file_data = source_zip.read(zip_item.filename)
                        
                        new_zip_item = zipfile.ZipInfo(zip_item.filename)
                        new_zip_item.date_time = zip_item.date_time
                        new_zip_item.compress_type = zip_item.compress_type
                        
                        # 设置文件权限
                        permission = 0o550 << 16 if zip_item.filename.endswith('/') else 0o440 << 16
                        new_zip_item.external_attr = permission
                        
                        target_zip.writestr(new_zip_item, file_data)
            
            os.replace(temp_file_path, whl_file_path)
            
        except Exception as error:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)
            raise error
    
    @staticmethod
    def _transfer_artifacts(temp_dist_dir, original_dist_dir):
        """将构建产物传输回原目录"""
        os.makedirs(original_dist_dir, exist_ok=True)
        
        for whl_file in glob.glob(os.path.join(temp_dist_dir, '*.whl')):
            target_file = os.path.join(original_dist_dir, os.path.basename(whl_file))
            shutil.copy2(whl_file, target_file)

    def execute_secure_build(self):
        """执行安全构建流程"""
        with tempfile.TemporaryDirectory(prefix="secure_build_") as temp_dir:
            logger.info("创建安全构建环境")
            
            self._copy_required_files(temp_dir)
            os.chdir(temp_dir)
            
            try:
                logger.info("开始构建wheel包")
                build_result = subprocess.run(
                    [sys.executable, 'setup.py', 'bdist_wheel'],
                    capture_output=True, text=True, check=True
                )
                
                if build_result.stderr:
                    logger.warning("构建过程中出现警告")
                
                # 在安全环境中修复权限
                temp_dist_dir = os.path.join(temp_dir, 'dist')
                if os.path.exists(temp_dist_dir):
                    logger.info("执行权限修复")
                    for whl_file in glob.glob(os.path.join(temp_dist_dir, '*.whl')):
                        SecureBuildManager._fix_whl_permissions(whl_file)
                
                # 传输构建产物
                original_dist_dir = os.path.join(self.source_dir, 'dist')
                SecureBuildManager._transfer_artifacts(temp_dist_dir, original_dist_dir)
                
                logger.info("安全构建完成")
                
            except subprocess.CalledProcessError as error:
                logger.error("构建过程失败")
                raise error
            finally:
                os.chdir(self.original_cwd)

    def _copy_required_files(self, target_dir):
        """复制必要的文件到目标目录"""
        required_items = ['setup.py', 'server']
        
        for item in required_items:
            source_path = os.path.join(self.source_dir, item)
            target_path = os.path.join(target_dir, item)
            
            if not os.path.exists(source_path):
                continue
                
            if os.path.isfile(source_path):
                shutil.copy2(source_path, target_path)
            else:
                shutil.copytree(source_path, target_path, dirs_exist_ok=True)


class BuildCleaner:
    """构建清理器"""
    
    def __init__(self, base_dir):
        self.base_dir = base_dir
    
    def clean_all(self):
        """清理所有构建产物"""
        cleanup_patterns = [
            'build',
            'dist',
            '*.egg-info',
            '__pycache__',
            '*.pyc',
            '*.pyo'
        ]
        
        for pattern in cleanup_patterns:
            for path in glob.glob(os.path.join(self.base_dir, pattern)):
                if not os.path.exists(path):
                    continue
                    
                try:
                    if os.path.isfile(path):
                        os.remove(path)
                    else:
                        shutil.rmtree(path)
                except Exception as error:
                    logger.warning(f"清理失败: {error}")


def main():
    """主执行函数"""
    if len(sys.argv) > 1 and sys.argv[1] == 'clean':
        cleaner = BuildCleaner(os.path.dirname(os.path.abspath(__file__)))
        cleaner.clean_all()
    else:
        source_directory = os.path.dirname(os.path.abspath(__file__))
        build_manager = SecureBuildManager(source_directory)
        build_manager.execute_secure_build()


if __name__ == "__main__":
    main()


# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

from mindspore import mint, ops, _no_grad
from mindspore import Tensor
from mindspore import dtype as mstype

from msprobe.core.common.log import logger


@_no_grad()
def square_sum(x: Tensor):
    return (x * x).sum()


@_no_grad()
def get_min(x: Tensor):
    return mint.min(x)


@_no_grad()
def get_mean(x: Tensor):
    return mint.mean(x.astype(mstype.float32))


@_no_grad()
def get_norm(x: Tensor):
    norm_func = mint.norm if hasattr(mint, "norm") else ops.norm
    return norm_func(x.astype(mstype.float32))


@_no_grad()
def get_max(x: Tensor):
    return mint.max(x)


@_no_grad()
def get_zeros(x: Tensor, eps: float):
    if x.numel() == 0:
        return Tensor(float('nan'))
    return mint.sum(mint.abs(x) < eps) / x.numel()


@_no_grad()
def get_nans(t):
    return ops.isnan(t.astype(mstype.float32)).sum()


def get_shape(t):
    return t.shape


def get_dtype(t):
    return t.dtype


FUNC_MAP = {
    "min": get_min,
    "max": get_max,
    "mean": get_mean,
    "norm": get_norm,
    "nans": get_nans,
    "zeros": get_zeros,
    "shape": get_shape,
    "dtype": get_dtype
}


def max_eigenvalue(input_tensor: Tensor, num_iterations=3):
    input_tensor = input_tensor.float()
    try:
        check_tensor_dim(input_tensor, 2)
    except (TypeError, ValueError) as e:
        logger.warning(f"calcute max eigenvalue failed, {e}")
        return Tensor(0)
    in_features = input_tensor.shape[1]
    u_tensor = ops.randn(in_features)
    u_norm = u_tensor.norm()
    if u_norm == 0:
        return Tensor(0)
    u_tensor /= u_tensor.norm()
    input_seq = ops.matmul(input_tensor.T, input_tensor)
    for _ in range(num_iterations):
        v_tensor = ops.matmul(input_seq, u_tensor)
        spectral_norm = ops.matmul(v_tensor.T, u_tensor)
        v_norm = v_tensor.norm()
        if v_norm > 0:
            u_tensor = v_tensor / v_norm
        else:
            spectral_norm = Tensor(0)
            break
    return spectral_norm.sqrt()


def check_tensor_dim(tensor, n):
    if not isinstance(tensor, Tensor):
        raise TypeError(
            f"Input must be a mindspore Tensor, but got {type(tensor)} instead."
            )
    if len(tensor.shape) < n:
        raise ValueError(
            f"tensor dim must be at least {n} dimensions."
            f"Got shape: {tuple(tensor.shape)} with {tensor.dim()} dims"
            )


def cal_entropy(qk_tensor: Tensor, mask=None):
    try:
        check_tensor_dim(qk_tensor, 2)
    except (TypeError, ValueError) as e:
        logger.warning(f"calculate entropy failed, {e}")
        return Tensor(0), Tensor(0)
    if mask is None:
        mask = ops.tril(ops.ones((qk_tensor.shape[1], qk_tensor.shape[1])))
    qk_tensor = qk_tensor - ops.amax(qk_tensor, axis=1, keepdims=True)
    qk_tensor = qk_tensor.masked_fill(mask == 0, float('-inf'))
    softmax_qkt = ops.softmax(qk_tensor.float(), axis=1)
    softmax_max = ops.mean(ops.amax(softmax_qkt, axis=1))
    entropy = ops.mean(-ops.nansum(softmax_qkt * ops.log(softmax_qkt), axis=1))
    return entropy, softmax_max


def cal_stable_rank(weight: Tensor):
    eig = max_eigenvalue(weight)
    if eig == Tensor(0):
        return Tensor(0), Tensor(0)
    f_norm = ops.norm(weight, ord='fro')
    return f_norm / eig, eig


def cal_qkt(q_h: Tensor, k_h: Tensor, order="s,b,h,d"):
    # q_h shape is (s, b, h, d)
    try:
        check_tensor_dim(q_h, 4)
        check_tensor_dim(k_h, 4)
    except (TypeError, ValueError) as e:
        logger.warning(f"calculatee qkt failed, {e}")
        return Tensor(0)
    if order == "s,b,h,d":
        qkt = ops.matmul(q_h[:, 0, 0, :], k_h[:, 0, 0, :].t()) / q_h.shape[-1] ** 0.5
    elif order == "b,s,h,d":
        qkt = ops.matmul(q_h[0, :, 0, :], k_h[0, :, 0, :].t()) / q_h.shape[-1] ** 0.5
    else:
        logger.warning(f"Calculate qk tensor failed: Order unsupported.")
        qkt = Tensor(0)
    return qkt



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
This file mainly involves vector comparison class definition.
"""
import copy
import os
import sys
import multiprocessing
import argparse
import csv
import time
import psutil

from msprobe.msaccucmp.dump_parse import dump, mapping
from msprobe.msaccucmp.vector_cmp.compare_detail import detail
from msprobe.msaccucmp.algorithm_manager.algorithm_manager import AlgorithmManager
from msprobe.msaccucmp.vector_cmp.fusion_manager import compare_result
from msprobe.msaccucmp.vector_cmp.fusion_manager.compare_rule import CompareRule
from msprobe.msaccucmp.format_manager.format_manager import FormatManager
from msprobe.msaccucmp.vector_cmp.fusion_manager.compare_fusion_op import FusionOpComparison
from msprobe.msaccucmp.vector_cmp.compare_detail.compare_detail import DetailComparison
from msprobe.msaccucmp.vector_cmp.compare_detail.compare_detail import DumpDetailComparison
from msprobe.msaccucmp.dump_parse.dump import DumpType
from msprobe.msaccucmp.cmp_utils import log, utils, utils_type, path_check
from msprobe.msaccucmp.cmp_utils.utils import safe_path_string
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.vector_cmp.range_manager.range_manager import RangeManager
from msprobe.msaccucmp.vector_cmp.range_manager.range_mode import RangeMode
from msprobe.msaccucmp.vector_cmp.range_manager.select_mode import SelectMode
from msprobe.msaccucmp.overflow.overflow_detection import OverflowDetection
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError
from msprobe.msaccucmp.cmp_utils.utils import sanitize_csv_value


class VectorComparison:
    """
    The class for vector compare
    """

    MULTI_THREAD_RESULT_COUNT = 3
    MULTI_THREAD_RETURN_CODE_INDEX = 0
    MULTI_THREAD_DUMP_MATCH_INDEX = 1
    MULTI_THREAD_COMPARE_RESULT_INDEX = 2
    MULTI_THREAD_MAX_NUM = 16

    def __init__(self: any, arguments: any = None) -> None:
        self.compare_rule = None
        self.compare_data = None
        self.detail_info = None
        self.format_manager = None
        self.args = {}
        if arguments:
            self._init_by_input_arguments(arguments)
        else:
            self._init_by_input_parse()

    @staticmethod
    def _parser_cmd(parse: any) -> None:
        parse.add_argument("-l", dest="left_dump_path", type=safe_path_string,
                           help="<Required> the left dump path, the data compared with golden data", required=True)
        parse.add_argument("-r", dest="right_dump_path", type=safe_path_string,
                           help="<Required> the right dump path, the golden data", required=True)
        parse.add_argument("-o", dest="output_path", help="<Required> output file path", type=safe_path_string,
                           required=True)
        parse.add_argument("-f", dest="fusion_json_file_path", default="", type=safe_path_string,
                           help="<Optional> fusion json file path")
        parse.add_argument("-q", dest="quant_fusion_rule_file_path", type=safe_path_string,
                           default="", help="<Optional> quant fusion rule file path")
        parse.add_argument("-d", dest="op_name", default="", help="<Optional> detail operator name", required=False)
        parse.add_argument("-t", dest="detail_type", default="output", required=False,
                           help="<Optional> detail type for operator, input or output, the default is output")
        parse.add_argument("-i", dest="detail_index", default="0",
                           help="<Optional> detail index for input or output, the default is 0", required=False)
        parse.add_argument("-csv", dest="csv", action="store_true",
                           default=False, help="<Optional> save file as csv format", required=False)
        parse.add_argument("-custom", dest="custom_path", default="", type=safe_path_string,
                           help="<Optional> user-defined path, including format conversion", required=False)
        parse.add_argument("-ffts", dest="ffts", action="store_true",
                           help="<optional> Enable the comparison between ffts+ and ffts+. "
                                "Direct comparison is performed without data combination. ")

    @staticmethod
    def _process_single_op_max_line_parameters(max_line: int) -> None:
        if max_line < ConstManager.DETAIL_LINE_COUNT_RANGE_MIN or max_line > ConstManager.DETAIL_LINE_COUNT_RANGE_MAX:
            log.print_out_of_range_error(None, '--max_line argument', max_line, '{} - {}'
                                         .format(ConstManager.DETAIL_LINE_COUNT_RANGE_MIN,
                                                 ConstManager.DETAIL_LINE_COUNT_RANGE_MAX))
            raise CompareError(CompareError.MSACCUCMP_INVALID_PARAM_ERROR)

    def set_output_path(self: any, output_path: str) -> None:
        """
        Set output path
        :param output_path: the output path
        """
        if os.path.islink(os.path.abspath(output_path)):
            log.print_error_log('The path "%r" is a softlink, not permitted.' % output_path)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)
        self.output_path = os.path.realpath(output_path)

    def check_arguments_valid(self: any) -> None:
        """
        Check arguments valid, if invalid, throw exception
        """
        self.compare_rule.check_arguments_valid()
        exist = False
        path_type = path_check.PathType.File
        if self.detail_info:
            exist = True
            path_type = path_check.PathType.Directory
            self.detail_info.check_arguments_valid()

        ret = path_check.check_output_path_valid(self.output_path, exist, path_type)
        if ret != CompareError.MSACCUCMP_NONE_ERROR:
            raise CompareError(ret)

        # delete old result
        if os.path.exists(self.output_path) and not self.detail_info:
            os.remove(self.output_path)
        self.compare_data.check_arguments_valid(self.compare_rule.fusion_json_file_path,
                                                self.compare_rule.quant_fusion_rule_file_path,
                                                self.compare_rule.close_fusion_rule_file_path)
        self._filter_left_dump_is_npy_overflow()
        self.format_manager.check_arguments_valid()

    def compare(self: any) -> int:
        """
        Compare for vector or detail
        """
        # 1. check arguments valid
        self.check_arguments_valid()
        # 2. parse json file
        self.compare_rule.parse_fusion_rule(self.compare_data)
        if ConstManager.RANGE_MANAGER_KEY in self.args:
            self.args.get(ConstManager.RANGE_MANAGER_KEY).check_input_valid(
                self.compare_rule.fusion_info.op_list[-1].attr.get_op_sequence())
        self.args["input_nodes"] = self.compare_rule.fusion_info.input_nodes
        # 3. do compare detail
        if self.detail_info:
            return self._compare_detail()
        # 4. do mapping
        if self.args.get("mapping"):
            return self._make_table()
        # 5. do compare vector
        return self._compare_vector()

    def _init_by_input_arguments(self, arguments) -> None:
        self.compare_rule = CompareRule(arguments.fusion_rule_file,
                                        arguments.quant_fusion_rule_file,
                                        arguments.close_fusion_rule_file)
        self.compare_data = dump.CompareData(
            os.path.realpath(arguments.my_dump_path),
            os.path.realpath(arguments.golden_dump_path),
            arguments.dump_version,
            arguments.ffts,
            arguments.fusion_rule_file)
        if arguments.op_name:
            self._process_single_op_parameters(arguments)
        else:
            self._process_output_path_parameter(arguments)
        self.args["csv"] = True
        self.format_manager = FormatManager(arguments.custom_script_path)
        self.args["algorithm_manager"] = AlgorithmManager(arguments.custom_script_path,
                                                          arguments.algorithm, arguments.algorithm_options)
        self.args["mapping"] = arguments.mapping
        self.args["overflow_detection"] = arguments.overflow_detection
        self.args["advisor"] = arguments.advisor
        self.args["input_nodes"] = []
        if arguments.range:
            self.args["range"] = arguments.range
            self.args[ConstManager.RANGE_MANAGER_KEY] = RangeMode(arguments.range)
        elif arguments.select:
            self.args["select"] = arguments.select
            self.args[ConstManager.RANGE_MANAGER_KEY] = SelectMode(arguments.select)
        self.args["my_dump_path"] = arguments.my_dump_path
        self.args["golden_dump_path"] = arguments.golden_dump_path
        self.args["max_cmp_size"] = arguments.max_cmp_size

    def _init_by_input_parse(self) -> None:
        parse = argparse.ArgumentParser()
        self._parser_cmd(parse)
        args, _ = parse.parse_known_args(sys.argv[1:])

        self.compare_rule = CompareRule(args.fusion_json_file_path,
                                        args.quant_fusion_rule_file_path)

        self.compare_data = dump.CompareData(os.path.realpath(args.left_dump_path),
                                             os.path.realpath(args.right_dump_path), ConstManager.OLD_DUMP_TYPE,
                                             args.ffts,
                                             args.fusion_json_file_path
                                             )
        if os.path.islink(os.path.abspath(args.output_path)):
            log.print_error_log('The path "%r" is a softlink, not permitted.' % args.output_path)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)
        self.output_path = os.path.realpath(args.output_path)
        if args.op_name:
            tensor_id = detail.TensorId(args.op_name, args.detail_type, args.detail_index)
            self.detail_info = detail.DetailInfo(tensor_id, ConstManager.DEFAULT_TOP_N, ignore_result=False,
                                                 max_line=ConstManager.MAX_DETAIL_INFO_LINE_COUNT)
        self.args["csv"] = args.csv
        self.format_manager = FormatManager(args.custom_path)
        self.args["algorithm_manager"] = AlgorithmManager('', 'all', '')
        self.args["mapping"] = False
        self.args["my_dump_path"] = args.left_dump_path
        self.args["golden_dump_path"] = args.right_dump_path

    def _check_both_dump_data(self: any) -> bool:
        both_dump_data = False
        if DumpType.Offline == self.compare_data.left_dump_info.type \
                and DumpType.Offline == self.compare_data.right_dump_info.type:
            both_dump_data = True
            if self.args.get("overflow_detection"):
                log.print_warn_log('Both compare data are NPU dump data, not support overflow detection.')
        return both_dump_data

    def _process_output_path_parameter(self: any, arguments: any) -> None:
        if arguments.mapping:
            file_name = 'mapping_%s.csv' % time.strftime("%Y%m%d%H%M%S", time.localtime(time.time()))
        else:
            file_name = 'result_%s.csv' % time.strftime("%Y%m%d%H%M%S", time.localtime(time.time()))
        if os.path.islink(os.path.abspath(arguments.output_path)):
            log.print_error_log('The path "%r" is a softlink, not permitted.' % arguments.output_path)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)
        self.output_path = os.path.join(os.path.realpath(arguments.output_path), file_name)

    def _process_single_op_parameters(self: any, arguments: any) -> None:
        tensor_type = ConstManager.OUTPUT
        tensor_index = '0'
        max_line = ConstManager.MAX_DETAIL_INFO_LINE_COUNT
        if arguments.max_line is not None:
            self._process_single_op_max_line_parameters(arguments.max_line)
            max_line = arguments.max_line
        if arguments.input:
            tensor_type = ConstManager.INPUT
            tensor_index = arguments.input
        if arguments.output:
            tensor_type = ConstManager.OUTPUT
            tensor_index = arguments.output
        if os.path.islink(os.path.abspath(arguments.output_path)):
            log.print_error_log('The path "%r" is a softlink, not permitted.' % arguments.output_path)
            raise CompareError(CompareError.MSACCUCMP_INVALID_PATH_ERROR)
        self.output_path = os.path.realpath(arguments.output_path)
        tensor_id = detail.TensorId(arguments.op_name, tensor_type, tensor_index)
        self.detail_info = detail.DetailInfo(tensor_id, arguments.topn, arguments.ignore_single_op_result, max_line)

    def _filter_left_dump_is_npy_overflow(self) -> None:
        """
        npy doesn't support overflow detection.
        We need turned the parameter 'overflow_detection' to False.
        Different types of dump files have different naming rules.
        The suffix of an npy file is '.npy'. Therefore, you only need to
        check the suffix of the path corresponding to any operator to
        determine whether the file is a npy file.
        If the parameter 'overflow_detection' is set to True, change it to False.
        """
        if self.args.get('overflow_detection') and self.compare_data.left_dump_info.op_name_to_file_map:
            file_path_list = []
            for _, file_path in self.compare_data.left_dump_info.op_name_to_file_map.items():
                file_path_list = file_path
                if file_path_list:
                    break
            if file_path_list:
                file_name = file_path_list[0]
                if file_name.endswith('.npy'):
                    self.args['overflow_detection'] = False

    def _compare_fusion_ops(self: any, fusion_op_names: list, lock: any) -> list:
        all_cmp_res = []

        _ = lock  # Bypassing parameter lock is not used
        for op_name in fusion_op_names:
            res = self._compare_by_fusion_op(op_name)
            all_cmp_res.append(res)
            # save result when 1000 operators are compared
        return all_cmp_res

    def _get_result_list(self, res):
        result = []
        for single_op_result in res[self.MULTI_THREAD_COMPARE_RESULT_INDEX]:
            for item in single_op_result.result_list:
                result.append(item)
        return result

    def _write_result_to_writer(self: any, result: list, output_file: any) -> None:
        for res in result:
            if len(res) != self.MULTI_THREAD_RESULT_COUNT:
                continue
            item_list = self._get_result_list(res)
            for item in item_list:
                if not item:
                    log.print_error_log("result item is empty, please check.")
                    raise CompareError(CompareError.MSACCUCMP_INDEX_OUT_OF_BOUNDS_ERROR)
                if self.args.get("csv"):
                    writer = csv.writer(output_file)
                    sanitized_item = [sanitize_csv_value(cell) for cell in item]
                    writer.writerow(sanitized_item)
                else:
                    item.pop()
                    each_row_str = " ".join(item)
                    output_file.write(f"\n{each_row_str}")

    def _save_cmp_result(self: any, result: list, lock: any) -> None:
        lock.acquire()
        try:
            with os.fdopen(os.open(self.output_path, ConstManager.WRITE_FLAGS,
                                   ConstManager.WRITE_MODES), 'a+', newline='') as output_file:
                self._write_result_to_writer(result, output_file)
        except IOError as io_error:
            log.print_open_file_error(self.output_path, io_error)
        finally:
            lock.release()

    def _compare_by_fusion_op(self: any, fusion_op_name: str) -> (int, bool, list):
        comparison = FusionOpComparison(fusion_op_name, self.compare_rule, self.compare_data, self.format_manager,
                                        self.args)
        return comparison.compare()

    def _get_max_process_num(self) -> int:
        if self.MULTI_THREAD_MAX_NUM == 1:
            return 1  # Bypassing test code entering `os.listdir`

        golden_dump_path = self.args.get("golden_dump_path")
        file_sizes = [1] + [os.path.getsize(os.path.join(golden_dump_path, ii)) for ii in os.listdir(golden_dump_path)]
        max_file_size = max(file_sizes)

        mem = psutil.virtual_memory()
        available_mem = mem.available
        mem_max_process_num = available_mem // max_file_size // 4

        cpu_max_process_num = int((multiprocessing.cpu_count() + 1) / 2)
        return min(mem_max_process_num, cpu_max_process_num, self.MULTI_THREAD_MAX_NUM)

    def _handle_multi_process(self: any, func: any, lock: any = None) -> list:
        # 2. compare operator by multi processes
        # 1 ensure multi processes number, which is half of the CPUs
        process_num = self._get_max_process_num()
        # 2 get all operator names
        if ConstManager.RANGE_MANAGER_KEY in self.args:
            all_op_names = self.args.get(ConstManager.RANGE_MANAGER_KEY).get_all_ops(self.compare_rule)
        else:
            all_op_names = self.compare_rule.fusion_info.fusion_op_name_to_op_map.keys()
        # 3 allocate all operator names evenly by multi processes number
        op_names = []
        for _ in range(process_num):
            op_names.append([])
        for i, op_name in enumerate(all_op_names):
            op_names[i % process_num].append(op_name)
        # 4 start multi processes, then waiting subprocess end of running
        all_task = []
        pool = multiprocessing.Pool(process_num)
        for fusion_op_names in op_names:
            if lock:
                task = pool.apply_async(func, args=(fusion_op_names, lock))
            else:
                task = pool.apply_async(func, args=(fusion_op_names,))
            all_task.append(task)
        pool.close()
        pool.join()
        return all_task

    def _compare_by_multi_process(self: any) -> (int, bool):
        # 1. write header to file
        if not self._write_header_to_file():
            return CompareError.MSACCUCMP_OPEN_FILE_ERROR, False
        # 2. compare operator by multi-processing
        all_task = self._handle_multi_process(self._compare_fusion_ops, multiprocessing.Manager().RLock())

        # 3. check subprocess return value
        ret = CompareError.MSACCUCMP_NONE_ERROR
        dump_match = False
        all_result = []
        result_mapping = {}
        cmp_res = []

        for task in all_task:
            all_result.extend(task.get())
        for res in all_result:
            if len(res) != self.MULTI_THREAD_RESULT_COUNT:
                continue
            if not ret:
                ret = res[self.MULTI_THREAD_RETURN_CODE_INDEX]
            if not dump_match:
                dump_match = res[self.MULTI_THREAD_DUMP_MATCH_INDEX]
            for single_op_cmp_res in res[self.MULTI_THREAD_COMPARE_RESULT_INDEX]:
                if single_op_cmp_res.is_ffts and not single_op_cmp_res.npu_vs_npu:
                    result_mapping[single_op_cmp_res.op_name] = single_op_cmp_res

        for i, res in enumerate(all_result):
            for single_op_cmp_res in res[self.MULTI_THREAD_COMPARE_RESULT_INDEX]:
                if single_op_cmp_res.is_ffts and single_op_cmp_res.op_name_origin_output_index_map:
                    single_op_cmp_res.find_pre_op(result_mapping)

            cmp_res.append(res)
            if (i + 1) % 1000 == 0:
                self._save_cmp_result(cmp_res, multiprocessing.Manager().RLock())
                cmp_res.clear()
        self._save_cmp_result(cmp_res, multiprocessing.Manager().RLock())
        return ret, dump_match

    def _compare_vector(self: any) -> int:
        ret, dump_match = self._compare_by_multi_process()
        utils.sort_result_file_by_index(self.output_path, self.args.get("csv"))
        if not dump_match:
            ret = CompareError.MSACCUCMP_NO_DUMP_FILE_ERROR
            if self.args.get("range"):
                log.print_warn_log('The model in [%s] range does not match the dump data.' % self.args.get('range'))
            elif self.args.get("select"):
                log.print_warn_log('The model in index list [%s] does not match '
                                   'the dump data.' % self.args.get('select'))
            else:
                log.print_error_log('The model does not match the dump data, '
                                    'please check the model and the dump data again.')
        else:
            if os.path.exists(self.output_path):
                log.print_write_result_info('comparison result', self.output_path)
                if self.args.get("advisor"):
                    self._do_advisor()
        return ret

    def _do_advisor(self):
        try:
            from msprobe.msaccucmp.advisor.compare_advisor import CompareAdvisor
        except ImportError as import_error:
            log.print_warn_log("Unable to import module: %s." % str(import_error))
            log.print_warn_log("Skip compare results Analysis.")
        else:
            out_path = os.path.dirname(self.output_path)
            compare_advisor = CompareAdvisor(self.output_path, self.args.get("input_nodes"), out_path)
            advisor_result = compare_advisor.advisor()
            message_list = advisor_result.print_advisor_log()
            advisor_result.gen_summary_file(out_path, message_list)

    def _compare_detail(self: any) -> int:
        """
        Compare detail by op name
        :return VectorComparisonErrorCode
        """
        if self.compare_rule.fusion_json_file_path == "" and self.compare_rule.quant_fusion_rule_file_path == "":
            log.print_warn_log('Both the offline fusion rule file path and '
                               'the quant fusion rule file path cannot be empty. '
                               'Please ensure that the data is reasonable.')
            if self.args.get("overflow_detection"):
                log.print_warn_log('Both compare data are NPU dump data, not support overflow detection.')
            comparison = DumpDetailComparison(self.detail_info, self.compare_data, self.output_path)
            return comparison.compare()
        if self.detail_info.tensor_id.op_name not in self.compare_rule.fusion_info.op_name_to_fusion_op_name_map:
            log.print_error_log('There is no "%s" in the fusion rule file.' % self.detail_info.tensor_id.op_name)
            return CompareError.MSACCUCMP_INVALID_PARAM_ERROR
        if self.args.get("overflow_detection") and not self._check_both_dump_data():
            overflow_detection = OverflowDetection(self.compare_data, self.detail_info.tensor_id.op_name)
            overflow_detection.process_op_overflow_detection()
        fusion_op_name = self.compare_rule.fusion_info.op_name_to_fusion_op_name_map.get(
            self.detail_info.tensor_id.op_name)
        fusion_op_comparison = FusionOpComparison(fusion_op_name, self.compare_rule, self.compare_data,
                                                  self.format_manager, self.args)
        comparison = DetailComparison(self.detail_info, fusion_op_comparison, self.output_path)
        return comparison.compare()

    def _write_header_to_file(self: any) -> bool:
        cur_op_header = self._pre_handle_header()
        try:
            with os.fdopen(os.open(self.output_path, ConstManager.WRITE_FLAGS,
                                   ConstManager.WRITE_MODES), 'a+', newline='') as output_file:
                header = compare_result.get_result_title(self.args.get('algorithm_manager'), cur_op_header,
                                                         self.args.get('overflow_detection'))
                if self.args.get("csv"):
                    writer = csv.writer(output_file)
                    sanitized_header = [sanitize_csv_value(cell) for cell in header]
                    writer.writerow(sanitized_header)
                else:
                    output_file.write(" ".join(header))
        except IOError as io_error:
            log.print_open_file_error(self.output_path, io_error)
            return False
        return True

    def _make_mapping_table_by_op_name(self: any, fusion_op_names: list) -> list:
        all_cmp_res = []
        for op_name in fusion_op_names:
            res = FusionOpComparison(op_name, self.compare_rule, self.compare_data, self.format_manager,
                                     self.args).make_gpu_and_npu_mapping_table()
            all_cmp_res += res
        return all_cmp_res

    def _make_table(self: any) -> int:
        all_task = self._handle_multi_process(self._make_mapping_table_by_op_name)
        origin_list = []
        for task in all_task:
            origin_list += task.get()
        origin_list.sort(key=lambda xx: int(xx[0]))
        try:
            with os.fdopen(os.open(self.output_path, ConstManager.WRITE_FLAGS, ConstManager.WRITE_MODES), 'a+',
                           newline='') as out_file:
                writer = csv.writer(out_file)
                header = ConstManager.MAPPING_FILE_HEADER
                RangeManager.adjust_header(header)
                sanitized_header = [sanitize_csv_value(cell) for cell in header]
                writer.writerow(sanitized_header)
                for item in origin_list:
                    sanitized_item = [sanitize_csv_value(cell) for cell in item]
                    writer.writerow(sanitized_item)
        except IOError as io_error:
            log.print_open_file_error(self.output_path, io_error)
            raise CompareError(CompareError.MSACCUCMP_OPEN_FILE_ERROR) from io_error
        if os.path.exists(self.output_path):
            log.print_write_result_info('mapping table result', self.output_path)
        return CompareError.MSACCUCMP_NONE_ERROR

    def _pre_handle_header(self: any) -> list:
        op_header = copy.deepcopy(ConstManager.VECTOR_COMPARE_HEADER)
        golden_dump_path = self.args.get("golden_dump_path")
        my_dump_path = self.args.get("my_dump_path")
        address_index = [i for i, x in enumerate(op_header) if x == 'Address']
        if utils.dump_path_contains_npy(golden_dump_path) and len(address_index) > 0:
            op_header.pop(address_index[-1])
        if utils.dump_path_contains_npy(my_dump_path) and len(address_index) > 0:
            op_header.pop(address_index[0])
        return op_header




# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

import json
from enum import Enum
from msprobe.visualization.utils import GraphConst, post_process_db_pbar
from msprobe.core.common.const import Const, CompareConst
from msprobe.core.common.log import logger
from msprobe.visualization.db_utils import DBActuator
from msprobe.visualization.graph.base_node import BaseNode
from msprobe.visualization.graph.node_op import NodeOp


class CommunicationType(Enum):
    """
    通信类型：发送、接收、发送接收
    """
    SEND = 'send'
    RECEIVE = 'receive'
    SEND_RECEIVE = 'send_receive'


class DistributedType(Enum):
    """
    分布式类型：点对点通信、集体通信
    """
    P2P = 'p2p'
    COLLECTIVE = 'collective'


CANNOT_MATCH = 'cannot match distributed node in rank'


class DistributedAnalyzer:

    def __init__(self, distributed_info: dict, overflow_check: bool):
        self.overflow_check = overflow_check
        self.config = {
            # 当前通信api名称: 匹配目标通信api名称, 获取rank信息的位置参数或关键字参数, 通信类型, 分布式类型
            'send': ['recv', GraphConst.DST, CommunicationType.SEND.value, DistributedType.P2P],
            'isend': ['irecv', GraphConst.DST, CommunicationType.SEND.value, DistributedType.P2P],
            'recv': ['send', GraphConst.SRC, CommunicationType.RECEIVE.value, DistributedType.P2P],
            'irecv': ['isend', GraphConst.SRC, CommunicationType.RECEIVE.value, DistributedType.P2P],
            'broadcast': ['broadcast', '1', CommunicationType.SEND.value, DistributedType.COLLECTIVE],
            'scatter': ['scatter', GraphConst.SRC, CommunicationType.SEND.value, DistributedType.COLLECTIVE],
            'gather': ['gather', GraphConst.DST, CommunicationType.RECEIVE.value, DistributedType.COLLECTIVE],
            'reduce': ['reduce', '1', CommunicationType.RECEIVE.value, DistributedType.COLLECTIVE]
        }
        self.group_node_mapping = {}
        self.distributed_info = distributed_info
        self._make_group_node_mapping()

    @staticmethod
    def _get_opposite_communication_type(action):
        if action == CommunicationType.SEND.value:
            return CommunicationType.RECEIVE.value
        elif action == CommunicationType.RECEIVE.value:
            return CommunicationType.SEND.value
        return action

    @staticmethod
    def _node_output_all_equal(data: dict, target_data: dict):
        keys_to_compare = [Const.DTYPE, Const.SHAPE, Const.MAX, Const.MIN, Const.MEAN, Const.NORM]
        return all(data.get(key) == target_data.get(key) for key in keys_to_compare)

    @staticmethod
    def _get_target_rank(node, rank, parameter):
        """
        点对点通信, 从输出数据参数src或dst, 获取通信目标rank
        一对多通信和多对一通信, 从输出数据参数src或dst或位置参数, 获取发送或接收的rank源头
        :param node: 当前节点
        :param rank: 当前rank
        :param parameter: 输出数据参数
        :return: 目标rank
        """
        target_rank = node.input_data.get(f'{node.id}{GraphConst.INPUT}{parameter}', {}).get('value')
        if target_rank is None:
            logger.debug(f'The parameter {parameter} of node {node.id} does not exist, {CANNOT_MATCH}{rank}')
        return target_rank

    @staticmethod
    def _get_group_info(node, rank):
        """
        获取当前通信节点的group参数中的group_ranks和group_id
        :param node: 当前通信节点
        :param rank: 当前rank
        :return: group_ranks和group_id
        """
        group = node.input_data.get(f'{node.id}{GraphConst.INPUT}group', {})
        if not group:
            logger.debug(f'The kwarg group of node {node.id} does not exist, {CANNOT_MATCH}{rank}')
            return None, None
        group_ranks = group.get('group_ranks')
        if not group_ranks:
            logger.debug(f'The group_ranks of node {node.id} does not exist, {CANNOT_MATCH}{rank}')
            return None, None
        group_ranks = DistributedAnalyzer._check_and_convert_group_ranks(group_ranks)
        if not group_ranks:
            return None, None
        group_id = group.get('group_id')
        if not group_id:
            logger.debug(f'The group_id of node {node.id} does not exist, {CANNOT_MATCH}{rank}')
            return None, None
        return group_ranks, group_id

    @staticmethod
    def _check_and_convert_group_ranks(group_ranks):
        if not isinstance(group_ranks, list):
            try:
                group_ranks = json.loads(group_ranks)
            except json.decoder.JSONDecodeError as e:
                logger.debug(f'Error occurred, group_rank cannot be converted to list: {e}')
                return []
        return group_ranks

    def distributed_match(self):
        for rank, nodes_dict in self.distributed_info.items():
            for node_id, node in nodes_dict.items():
                # 已经匹配过了
                if node.matched_distributed:
                    continue
                api_name, distributed_type = self._get_distributed_name_and_type(node_id)
                if api_name == GraphConst.BATCH_P2P:
                    self._batch_p2p_match(node, rank)
                elif distributed_type == DistributedType.P2P:
                    self._p2p_match(node, rank, api_name)
                else:
                    self._collective_match(node, rank, api_name)

    def _make_group_node_mapping(self):
        """
        建立通信节点的全局唯一标识映射
        key: rank号, value: unique_group_id与node_id之间的映射
        {
            "0": {
                "unique_group_id1": "node_id1",
                "unique_group_id2": "node_id2",
                "node_id1": "unique_group_id1",
                "node_id2": "unique_group_id2"
            },
            "1": {},
            "2": {}
        }
        """
        for rank, nodes_dict in self.distributed_info.items():
            group_count = {}
            group_info = {}
            batch_p2p_count = {}
            for node_id, node in nodes_dict.items():
                api_name, distributed_type = self._get_distributed_name_and_type(node_id)
                if api_name == GraphConst.BATCH_P2P:
                    self._make_batch_p2p_mapping(node, rank, batch_p2p_count)
                    continue
                elif distributed_type == DistributedType.P2P:
                    config_info = self.config.get(api_name)
                    target_rank = self._get_target_rank(node, rank, config_info[1])
                    if target_rank is None:
                        continue
                    # p2p通信节点，api名称+传输目标rank作为group_id
                    group_id = api_name + Const.RANK + str(target_rank)
                else:
                    # 其他通信节点直接获取group_id, 并拼接api名称
                    _, group_id = self._get_group_info(node, rank)
                    if not group_id:
                        continue
                    group_id += api_name
                # 同group_id的调用次数累计
                group_count[group_id] = group_count.get(group_id, 0) + 1
                # group_id+同group_id的调用次数作为唯一的unique_group_id
                unique_group_id = group_id + Const.REPLACEMENT_CHARACTER + str(group_count.get(group_id))
                group_info[unique_group_id] = node_id
                group_info[node_id] = unique_group_id
            if rank not in self.group_node_mapping:
                self.group_node_mapping[rank] = {}
            self.group_node_mapping[rank].update(group_info)

    def _make_batch_p2p_mapping(self, node, rank, batch_p2p_count):
        """
        给batch_isend_irecv接口的每个p2p内容赋予唯一标识
        """
        if rank not in self.group_node_mapping:
            self.group_node_mapping[rank] = {}
        params = []
        for info_dict in node.batch_p2p_info:
            op = info_dict.get(GraphConst.OP)
            target_rank = info_dict.get(GraphConst.PEER)
            if op is None or target_rank is None:
                logger.debug('Cannot get param op or peer.')
                continue
            group_id = op + Const.REPLACEMENT_CHARACTER + Const.RANK + str(target_rank) + \
                       Const.REPLACEMENT_CHARACTER + info_dict.get(GraphConst.GROUP_ID, '')
            batch_p2p_count[group_id] = batch_p2p_count.get(group_id, 0) + 1
            # 例如: isend_rank0_5a4d31ad765260ba50eb190f1f9fd163_1
            unique_group_id = group_id + Const.REPLACEMENT_CHARACTER + str(batch_p2p_count.get(group_id))
            params.append(unique_group_id)
            self.group_node_mapping.get(rank)[unique_group_id] = node.id
        if params:
            self.group_node_mapping.get(rank)[node.id] = params

    def _get_distributed_name_and_type(self, node_id):
        if Const.SEP not in node_id:
            raise ValueError(f'Invalid node id {node_id}.')
        api_name = node_id.split(Const.SEP)[1]
        if api_name in self.config:
            return api_name, self.config.get(api_name)[3]
        return api_name, DistributedType.COLLECTIVE

    def _get_target_node(self, rank, unique_group_id, api_name, target_rank, target_api_name=None):
        """
        获取名称匹配上的目标节点
        :param rank: 当前rank
        :param unique_group_id: 当前节点唯一group id
        :param api_name: 当前节点的api名称, 例如Distributed.isend.0.forward, api名称为isend
        :param target_rank: 与当前节点产生通信的rank
        :param target_api_name: 与当前节点产生通信的节点api名称, 仅p2p通信需要配置
        :return: 目标节点
        """
        target_nodes_dict = self.distributed_info.get(str(target_rank))
        if not target_nodes_dict:
            logger.debug(f'Node data does not exist, {CANNOT_MATCH}{target_rank}')
            return None
        target_group_mapping = self.group_node_mapping.get(str(target_rank))
        # p2p通信，想要获取目标节点，需要替换unique_group_id中的rank和api name,
        # 例如isend发送到rank1，对应的irecv接收自rank0, isend_rank1与irecv_rank0对应
        target_unique_group_id = (unique_group_id
                                  .replace(Const.RANK + str(target_rank), Const.RANK + str(rank))
                                  .replace(api_name, target_api_name)) if target_api_name else unique_group_id
        target_node_id = target_group_mapping.get(target_unique_group_id, '')
        target_node = target_nodes_dict.get(target_node_id)
        if not target_node:
            logger.debug(f'Node {target_node_id} does not exist, {CANNOT_MATCH}{target_rank}')
            return None
        return target_node

    def _add_node_matched_distributed(self, node, target_node, api_name, target_rank, reversal_type=False):
        """
        给当前节点添加matched_distributed字段信息
        :param node: 当前节点
        :param target_node: 匹配上的目标节点
        :param api_name: 当前节点的api名称
        :param target_rank: 匹配上的目标rank
        :param reversal_type: 是否需要反转通信类型，例如broadcast在rank0通信类型是发送，但在其他rank通信类型是接收
        """
        communications_type = self.config.get(api_name)[2]
        communications_type = self._get_opposite_communication_type(communications_type) if reversal_type \
            else communications_type
        index = target_node.data.get(GraphConst.OVERFLOW_LEVEL, CompareConst.NAN) if self.overflow_check \
            else target_node.data.get(GraphConst.JSON_INDEX_KEY, CompareConst.NAN)
        index = CompareConst.NAN if index is None else index
        matched_distributed = {
            'communications_type': communications_type,
            'nodes_info': {target_rank: [str(index), target_node.id]}
        }
        node.matched_distributed = matched_distributed

    def _p2p_match(self, node, rank, api_name):
        """
        点对点通信匹配

        根据当前点对点通信节点的输出数据中的src或dst参数, 确定目标rank, 并从目标rank中找到对应的点对点通信节点, 校验输出数据是否一致，
        校验通过则在两个匹配节点增加匹配信息
        Args:
            node: 当前点对点通信节点
            rank: 当前节点所属rank
            api_name: 当前节点的api名称
        Returns:
        """
        config_info = self.config.get(api_name)
        target_api_name = config_info[0]
        target_rank = self._get_target_rank(node, rank, config_info[1])
        if target_rank is None:
            return
        unique_group_id = self.group_node_mapping.get(str(rank), {}).get(node.id, '')
        target_node = self._get_target_node(rank, unique_group_id, api_name, target_rank, target_api_name)
        if not target_node:
            return
        target_config_info = self.config.get(target_api_name)
        source_rank = (target_node.input_data.get(f'{target_node.id}{GraphConst.INPUT}{target_config_info[1]}', {})
                       .get('value'))
        if source_rank is None:
            logger.debug(
                f'The kwarg {target_config_info[1]} of node {target_node.id} does not exist, '
                f'{CANNOT_MATCH}{source_rank}')
            return
        if source_rank != rank:
            # 点对点通信，待匹配目标节点包含的rank信息要与当前rank一致
            logger.debug(
                f'{node.id} of rank{rank} is expected to communicate with {target_node.id} of rank{target_rank}, '
                f'but the data shows that {target_node.id} communicates with rank{source_rank}.'
                f'The rank is inconsistent, cannot match distributed node')
            return

        self._add_node_matched_distributed(node, target_node, api_name, target_rank)
        self._add_node_matched_distributed(target_node, node, target_api_name, rank)

    def _collective_match(self, node, rank, api_name):
        """
        集体通信匹配

        一对多通信和多对一通信, 需要先获取节点输出数据中的src或dst或位置参数, 确定发送源或接收源, 多对多通信不需要
        :param node: 当前集体通信节点
        :param rank: 当前节点所属rank
        :param api_name: 当前节点的api名称
        :return:
        """
        communications_type = CommunicationType.SEND_RECEIVE.value
        config_info = self.config.get(api_name)
        if config_info:
            # 此时为一对多通信或多对一通信
            source_rank = self._get_target_rank(node, rank, config_info[1])
            if source_rank is None or str(source_rank) != str(rank):
                return
            communications_type = config_info[2]
        group_ranks, group_id = self._get_group_info(node, rank)
        if not group_ranks or not group_id:
            return
        group_ranks = DistributedAnalyzer._check_and_convert_group_ranks(group_ranks)
        if not group_ranks:
            return
        unique_group_id = self.group_node_mapping.get(str(rank), {}).get(node.id, '')
        matched_distributed = {'communications_type': communications_type}
        nodes_info = {}
        for target_rank in group_ranks:
            if str(target_rank) == str(rank):
                continue
            target_node = self._get_target_node(rank, unique_group_id, api_name, target_rank)
            if not target_node:
                continue
            _, target_group_id = self._get_group_info(target_node, target_rank)
            if not target_group_id:
                continue
            if group_id != target_group_id:
                logger.debug(
                    f'{node.id} of rank{rank} is expected to communicate with {target_node.id} of rank{target_rank}'
                    f', but the data shows that the group id of the two nodes are different, '
                    f'cannot match distributed node')
                continue
            # 给当前通信节点添加matched_distributed字段信息
            index = target_node.data.get(GraphConst.OVERFLOW_LEVEL, CompareConst.NAN) if self.overflow_check \
                else target_node.data.get(GraphConst.JSON_INDEX_KEY, CompareConst.NAN)
            index = CompareConst.NAN if index is None else index
            nodes_info[target_rank] = [str(index), target_node.id]
            if config_info:
                # 给匹配上的目标节点也添加matched_distributed字段信息
                self._add_node_matched_distributed(target_node, node, api_name, rank, True)
        if nodes_info:
            matched_distributed['nodes_info'] = nodes_info
            node.matched_distributed = matched_distributed

    def _batch_p2p_match(self, node, rank):
        """
        批量点对点匹配

        针对torch.distributed.batch_isend_irecv接口，其入参是一个包含点对点通信信息的集合，需要遍历集合对每个点对点通信信息进行匹配
        :param node: 当前集体通信节点
        :param rank: 当前节点所属rank
        :return:
        """
        unique_group_ids = self.group_node_mapping.get(str(rank), {}).get(node.id)
        if not unique_group_ids:
            return
        matched_distributed = [] if len(unique_group_ids) > 1 else {}
        for unique_group_id in unique_group_ids:
            try:
                id_info = unique_group_id.split(Const.REPLACEMENT_CHARACTER)
                api_name = id_info[0]
                target_api_name = self.config.get(api_name)[0]
                target_rank = int(id_info[1].replace(Const.RANK, ''))
            except Exception as e:
                logger.debug(f'Failed to parse batch p2p parameter with error info: {e}.')
                continue
            target_node = self._get_target_node(rank, unique_group_id, api_name, target_rank, target_api_name)
            if not target_node:
                continue
            communications_type = self.config.get(api_name)[2]
            index = target_node.data.get(GraphConst.OVERFLOW_LEVEL, CompareConst.NAN) if self.overflow_check \
                else target_node.data.get(GraphConst.JSON_INDEX_KEY, CompareConst.NAN)
            index = CompareConst.NAN if index is None else index
            matched_info = {
                'communications_type': communications_type,
                'nodes_info': {target_rank: [str(index), target_node.id]}
            }
            matched_distributed.append(matched_info) if isinstance(matched_distributed, list) \
                else matched_distributed.update(matched_info)
        if matched_distributed:
            node.matched_distributed = matched_distributed


def distributed_analyse(db_path: str, overflow_check: bool, pbar_info=None):
    distributed_info_data_source = {}
    db_actuator = DBActuator(db_path)
    rows = db_actuator.query_distributed_nodes_info()
    if not rows:
        if pbar_info:
            post_process_db_pbar(pbar_info, True)
        return
    logger.info('Start analyzing distributed nodes...')
    for row in rows:
        if len(row) < 7:
            continue
        node_id = row[0]
        input_data = row[1]
        rank = row[2]
        step = row[3]
        data_source = row[4]
        precision_index = row[5]
        overflow_level = row[6]
        node = BaseNode(NodeOp.function_api, node_id)
        node.data[GraphConst.JSON_INDEX_KEY] = precision_index
        node.data[GraphConst.OVERFLOW_LEVEL] = overflow_level
        try:
            node.input_data = json.loads(input_data)
        except json.JSONDecodeError as e:
            logger.debug(f"JSON parsing of input data for node {node_id} failed: {e}, data: {input_data}")
            continue

        # data_source -> step -> rank -> {node_id: node} 分类
        distributed_info_data_source.setdefault(data_source, {}) \
            .setdefault(step, {}) \
            .setdefault(str(rank), {})[node_id] = node

    for data_source, distributed_info_steps in distributed_info_data_source.items():
        for step, distributed_info in distributed_info_steps.items():
            analyzer = DistributedAnalyzer(distributed_info, overflow_check)
            analyzer.distributed_match()
            db_actuator.update_matched_distributed(analyzer.distributed_info, step, data_source)

    if pbar_info:
        post_process_db_pbar(pbar_info, True)
    logger.info('Successfully analyzed distributed nodes.')



# coding=utf-8
# -------------------------------------------------------------------------
#  This file is part of the MindStudio project.
# Copyright (c) 2025 Huawei Technologies Co.,Ltd.
#
# MindStudio is licensed under Mulan PSL v2.
# You can use this software according to the terms and conditions of the Mulan PSL v2.
# You may obtain a copy of Mulan PSL v2 at:
#
#          http://license.coscl.org.cn/MulanPSL2
#
# THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
# EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
# MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
# See the Mulan PSL v2 for more details.
# -------------------------------------------------------------------------

"""
Function:
Pytorch CompareData class. This class mainly involves the function of parse dump_data.
"""
import numpy as np

from msprobe.msaccucmp.cmp_utils import utils_type
from msprobe.msaccucmp.cmp_utils import log
from msprobe.msaccucmp.pytorch_cmp import hdf5_parser
from msprobe.msaccucmp.cmp_utils.constant.const_manager import ConstManager
from msprobe.msaccucmp.cmp_utils.constant.compare_error import CompareError


class DataType:
    """
    The class for pytorch dump data DataType
    """
    data_type = {
        "Float": 1,
        "Byte": 2,
        "Char": 3,
        "Short": 5,
        "Int": 6,
        "Long": 7,
        "Bool": 9,
        "Half": 10,
        "Double": 11
    }
    equivalent_type = [
        ("Int", "Long"),
        ("Float", "Half")
    ]

    @classmethod
    def get_name(cls: any, value: int) -> str:
        """
        Get the type name by the type value.
        :value: the code of data type
        """
        return list(DataType.data_type.keys())[list(DataType.data_type.values()).index(value)]

    @classmethod
    def get_value(cls: any, name: str) -> int:
        """
        Get the type value by the type name.
        :name: the name of data type
        """
        return DataType.data_type.get(name)


class CompareMap:
    """
    The class is used for op mapping.
    """
    def __init__(self: any) -> None:
        self.op_map = {}
        self.param_map = {}
        self._init_map_table()

    def get_mapping_opname(self: any, opname: str) -> list:
        """
        Get mapping op name in map table.
        :opname: the op name before mapping
        """
        if opname in list(self.op_map.keys()):
            return self.op_map.get(opname)
        return []

    def get_mapping_param(self: any, opname: str, param: str) -> str:
        """
        Get mapping param in map table.
        :opname: the param name before mapping
        """
        if opname not in list(self.param_map.keys()):
            return ''
        item = self.param_map.get(opname)
        if param in item.keys():
            return item[param]
        return ''

    def get_mapping_opname_all(self: any) -> list:
        """
        Get all the mapped opname.
        """
        return list(self.op_map.values())

    def _init_map_table(self: any) -> None:
        self.op_map = {
            'CudnnBatchNormBackward': ['NativeBatchNormBackward'],
            'NativeBatchNormBackward': ['CudnnBatchNormBackward'],
            'CudnnConvolutionBackward': ['NpuConvolutionBackward'],
            'ThnnConvDepthwise2DBackward': ['NpuConvolutionBackward'],
            'NpuConvolutionBackward': ['CudnnConvolutionBackward', 'ThnnConvDepthwise2DBackward']
        }

        self.param_map = {
            'CudnnBatchNormBackward': {'epsilon': 'eps'},
            'NativeBatchNormBackward': {'eps': 'epsilon'},
            'CudnnConvolutionBackward': {'self': 'input'},
            'ThnnConvDepthwise2DBackward': {'self': 'input'},
            'NpuConvolutionBackward': {'input': 'self'}
        }


class CompareData:
    """
    The class for compare data, left dump data and right dump data.
    """
    NPU_PREFIX = "Npu"
    GPU_PREFIX = "Cudnn"
    CPU_PREFIX = "Thnn"

    def __init__(self: any, my_dump_path: str, golden_dump_path: str) -> None:
        self.mapping = CompareMap()
        self.my_dump = hdf5_parser.Hdf5Parser(my_dump_path, hdf5_parser.Hdf5Parser.MY_DUMP_FILE,
                                              self.mapping.get_mapping_opname_all())
        self.golden_dump = hdf5_parser.Hdf5Parser(golden_dump_path, hdf5_parser.Hdf5Parser.GOLDEN_DUMP_FILE,
                                                  self.mapping.get_mapping_opname_all())
        self.orders_num = 0

    @staticmethod
    def get_original_opname(ext_opname: str) -> str:
        """
        Get original op name in ext_opname.
        :ext_opname: extend op name. such as cov2d:2
        """
        if ConstManager.DELIMITER not in ext_opname:
            log.print_error_log(f": not in {ext_opname}, please check.")
            raise CompareError(CompareError.MSACCUCMP_NAME_ERROR)
        op_name, _ = ext_opname.split(ConstManager.DELIMITER, 1)
        return op_name

    @staticmethod
    def _is_equivalent_type(my_dump_data_type: int, golden_dump_data_type: int, type_info: str) -> bool:
        equivalent_type = set()
        for item in DataType.equivalent_type:
            equivalent_type.clear()
            for type_name in item:
                equivalent_type.add(DataType.get_value(type_name))
            if my_dump_data_type in equivalent_type \
                    and golden_dump_data_type in equivalent_type:
                message = 'The DataType on both sides are Compatible, {}'.format(type_info)
                log.print_info_log(message)
                return True
        return False

    @staticmethod
    def _check_stride(dataset_path: str, shape: list, stride: list) -> bool:
        message = "shape={},stride={},stride is invalid". \
            format(tuple(shape), tuple(stride))

        if len(shape) != len(stride):
            log.print_warn_log("[{}]:{}".format(dataset_path, message))
            return False

        expect_data_num = 0
        real_data_num = 1
        for (index, _) in enumerate(shape):
            expect_data_num += ((shape[index] - 1) * stride[index])
            real_data_num *= shape[index]
        expect_data_num += 1
        if expect_data_num > real_data_num:
            log.print_warn_log("[{}]:{}".format(dataset_path, message))
            return False
        return True

    def get_golden_dataset(self: any, ext_opname: str, my_dump_dataset_path: str) -> (bool, str, str):
        """
        Get the golden_dataset_path that matches the my_dump_dataset_path.
        :ext_opname: the extend op name. such as cov2d:2
        :my_dump_dataset_path: the my dump dataset path that get from dump data
        """
        message = 'No data match with {} in golden dump data.' \
            .format(my_dump_dataset_path)
        if ext_opname in self.golden_dump.ext_opname_dataset_map.keys():
            golden_ext_opname = ext_opname
        else:
            golden_ext_opname = self._opname_map(ext_opname, self.golden_dump.device_type)
            if golden_ext_opname not in self.golden_dump.ext_opname_dataset_map.keys():
                return False, '', message

        golden_dataset_path = self._construct_dataset_path(
            ext_opname, my_dump_dataset_path, self.golden_dump, golden_ext_opname)
        if golden_dataset_path and self.golden_dump.have_dataset(golden_ext_opname, golden_dataset_path):
            return True, golden_dataset_path, ''
        log.print_warn_log(message)
        return False, '', message

    def get_dump_data(self: any, my_dataset_path: str, golden_dataset_path: str) -> (any, any, str):
        """
        Get my dump data and golden dump data.
        :my_dataset_path: my dataset path in my dump data
        :golden_dataset_path: golden dataset path in golden dump data
        """
        success, message = self._check_data_type(my_dataset_path, golden_dataset_path)
        if not success:
            log.print_warn_log(message)
            raise CompareError(CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR, message)

        my_dump_data = self.my_dump.get_dump_data(my_dataset_path)
        golden_dump_data = self.golden_dump.get_dump_data(golden_dataset_path)
        converted_golden_dump_data = self._converted_stride(
            golden_dump_data, golden_dataset_path)
        return my_dump_data, converted_golden_dump_data, ''

    def get_not_matched_golden_datasets(self: any) -> list:
        """
        Find all the dataset that does not match with my dump in the golden file.
        """
        not_matched_info = []
        # Unmatched items. To display them at the end of the result file after sorting,
        # order must be the maximum value plus 1.
        for golden_ext_opname, _ in self.golden_dump.ext_opname_dataset_map.items():
            all_not_matched = True
            failed_info = []
            golden_dataset = ''
            for golden_dataset in self.golden_dump.ext_opname_dataset_map[golden_ext_opname]:
                if golden_ext_opname in self.my_dump.ext_opname_dataset_map.keys():
                    my_dump_ext_opname = golden_ext_opname
                else:
                    my_dump_ext_opname = self._opname_map(golden_ext_opname)
                if not self._reverse_match_non_load_mode(my_dump_ext_opname, golden_dataset, failed_info):
                    continue
                if not self._reverse_match_process(my_dump_ext_opname, golden_ext_opname,
                                                   golden_dataset, failed_info):
                    continue
                all_not_matched = False

            if all_not_matched:
                failed_info.clear()
                _, op_name, order, _ = golden_dataset.split('/', 3)
                message = 'No data match with /{}/{} in my dump data.'.format(op_name, order)
                failed_info.append([max(self.my_dump.get_all_orders()) + 1, 'NaN',
                                    '/{}/{}'.format(op_name, order), message])
            not_matched_info.extend(failed_info)

        return not_matched_info

    def set_compare_input_flag(self: any) -> None:
        """
        Sets whether to compare inputs.
        """
        if self.my_dump.need_compare_input:
            self.golden_dump.need_compare_input = True

    def check_my_dump_file_valid(self: any) -> None:
        """
        Check whether the device type is NPU.
        """
        if not self.my_dump.ext_opname_dataset_map:
            log.print_warn_log('my dump file is empty!')
            raise CompareError(
                CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR)

        if self.my_dump.device_type != utils_type.DeviceType.NPU.value:
            log.print_error_log('My dump file is not the dump data of the model'
                                ' executed on the AI processor, please check -m param!')
            raise CompareError(
                CompareError.MSACCUCMP_INVALID_DUMP_DATA_ERROR)

    def parse_dump_file(self: any) -> int:
        """
        Parsing the data to be compared.
        """
        ret = self.my_dump.parse_dump_file()
        if ret == CompareError.MSACCUCMP_NONE_ERROR:
            self.check_my_dump_file_valid()
            self.set_compare_input_flag()
            self.orders_num = len(self.my_dump.get_all_orders())
            return self.golden_dump.parse_dump_file()
        return ret

    def open_file(self: any, model: str) -> None:
        """
        Open the my dump and golden dump files.
        """
        self.my_dump.open_file(model)
        self.golden_dump.open_file(model)

    def close_file(self: any) -> None:
        """
        Close the my dump and golden dump files.
        """
        self.my_dump.close_file()
        self.golden_dump.close_file()

    def get_all_orders(self: any) -> list:
        """
        Get all orders incrementally.
        """
        return sorted(self.my_dump.get_all_orders())

    def get_ext_opname_by_order(self: any, order: int) -> list:
        """
        Get the list of extend op name corresponding to the order.
        :order: the order of Op execution
        """
        return self.my_dump.get_ext_opname_group_by_order(order)

    def get_my_dump_datasets(self: any, ext_opname: str) -> list:
        """
        Get the dump datasets(include input and output) by ext_opname.
        :ext_opname: the extend op name. such as cov2d:2
        """
        if ext_opname in self.my_dump.ext_opname_dataset_map.keys():
            return self.my_dump.ext_opname_dataset_map[ext_opname]
        return []

    def _construct_dataset_path(self: any, src_ext_opname: str, src_dataset_path: str, dst_dump_data: any,
                                dst_ext_opname: str) -> str:
        """
        Construct the dataset path based on the peer path.
        :ext_opname: the extend op name. such as cov2d:2
        :dataset_path: the base dataset_path
        :gen_golden_path: if the Dataset is the expect type(golden)
        """
        dst_order = dst_dump_data.get_order_by_ext_opname(dst_ext_opname)

        # e.g. dataset_path:/AddmmBackward/4/input/mat1,dataset_name: input/mat1
        _, _, src_order, direction, src_dataset_name = src_dataset_path.split('/', 4)

        # in load comparison mode, inputs do not need to be compared, and
        # the dataset paths of my dump and golden dump must be the same order.
        if dst_dump_data.is_load_mode():
            dst_order = src_order

        dst_opname = self.get_original_opname(dst_ext_opname)

        dst_dataset_path = "/{}/{}/{}/{}".format(dst_opname, dst_order, direction, src_dataset_name)
        if dst_dump_data.have_dataset(dst_ext_opname, dst_dataset_path):
            return dst_dataset_path

        src_opname = self.get_original_opname(src_ext_opname)
        # get param name at suffix
        src_param_name = src_dataset_path.split('/')[-1]
        mapping_param_name = self.mapping.get_mapping_param(src_opname, src_param_name)
        if mapping_param_name:
            # replace param name at suffix
            return "{}{}".format(dst_dataset_path[:-len(src_param_name)], mapping_param_name)

        return ''

    def _opname_map_by_map_table(self: any, ext_opname: str, device_type: int) -> str:
        replaced_ext_opname = ''
        opname = self.get_original_opname(ext_opname)
        mapping_opname_list = self.mapping.get_mapping_opname(opname)
        if not mapping_opname_list:
            return replaced_ext_opname
        for mapping_opname in mapping_opname_list:
            replaced_ext_opname = ext_opname.replace(opname, mapping_opname, 1)
            if device_type == utils_type.DeviceType.NPU.value:
                if replaced_ext_opname in self.my_dump.ext_opname_dataset_map.keys():
                    return replaced_ext_opname
            else:
                if replaced_ext_opname in self.golden_dump.ext_opname_dataset_map.keys():
                    return replaced_ext_opname
        return ''

    def _opname_map(self: any, ext_opname: str, device_type: int = utils_type.DeviceType.NPU.value) -> str:
        """
        Processes mappings between NPUs, GPNs, and CPU operators.
        :ext_opname: the extend op name. such as cov2d:2
        :device_type: the device_type at my dump side
        """
        mapped_opname = self._opname_map_by_map_table(ext_opname, device_type)
        if mapped_opname:
            return mapped_opname

        if ext_opname.startswith(self.NPU_PREFIX):
            if device_type == utils_type.DeviceType.CPU.value:
                return ext_opname.replace(self.NPU_PREFIX, self.CPU_PREFIX, 1)
            if device_type == utils_type.DeviceType.GPU.value:
                return ext_opname.replace(self.NPU_PREFIX, self.GPU_PREFIX, 1)
        # reverse lookup
        if ext_opname.startswith(self.CPU_PREFIX):
            return ext_opname.replace(self.CPU_PREFIX, self.NPU_PREFIX, 1)
        if ext_opname.startswith(self.GPU_PREFIX):
            return ext_opname.replace(self.GPU_PREFIX, self.NPU_PREFIX, 1)
        return ext_opname

    def _check_data_type(self: any, my_dataset_path: str, golden_dataset_path: str) -> (bool, str):
        """
        Check the DataType of my dump data and golden dump data.
        :my_dataset_path: the my dump dataset path
        :golden_dataset_path: the golden dump dataset path
        """
        message = "my_dataset_path is {}, golden_dataset_path" \
                  " is {}.".format(my_dataset_path, golden_dataset_path)
        attr_ok, my_dump_data_type = self.my_dump.get_dump_data_attr(
            my_dataset_path, utils_type.DatasetAttr.DataType.name)
        if not attr_ok:
            return False, "Get the attr 'DataType' of {} failed! {}".format("my_dump", message)

        attr_ok, golden_dump_data_type = self.golden_dump.get_dump_data_attr(
            golden_dataset_path, utils_type.DatasetAttr.DataType.name)
        if not attr_ok:
            return False, "Get the attr 'DataType' of {} failed! {}".format("golden_dump", message)

        type_info = 'my dump data path is {}, the data type of both sides is({},{}).' \
            .format(my_dataset_path, DataType.get_name(my_dump_data_type),
                    DataType.get_name(golden_dump_data_type))
        if my_dump_data_type == golden_dump_data_type:
            return True, ''

        if self._is_equivalent_type(my_dump_data_type, golden_dump_data_type, type_info):
            return True, ''

        message = 'The DataType on both sides are different, {}'.format(type_info)
        return False, message

    def _converted_stride(self: any, dump_data: any, dataset_path: str) -> any:
        # convert GPU/CPU stride
        attr_ok, device_type = self.golden_dump.get_dump_data_attr(dataset_path,
                                                                   utils_type.DatasetAttr.DeviceType.name)
        if attr_ok and device_type in (utils_type.DeviceType.GPU.value, utils_type.DeviceType.CPU.value):
            have_stride_attr, stride_attr = self.golden_dump.get_dump_data_attr(dataset_path,
                                                                                utils_type.DatasetAttr.Stride.name)
            if have_stride_attr and self._check_stride(dataset_path, list(dump_data.shape), list(stride_attr)):
                dump_data_flatten = dump_data.flatten()
                real_stride = (dump_data_flatten.strides[0] * i for i in list(stride_attr))
                return np.lib.stride_tricks.as_strided(
                    dump_data_flatten, shape=dump_data.shape, strides=real_stride)
        return dump_data

    def _reverse_match_non_load_mode(self: any, my_dump_ext_opname: str, golden_dataset: str,
                                     failed_info: list) -> bool:
        message = 'No data match with {} in my dump data.'.format(golden_dataset)
        if not self.my_dump.is_load_mode() \
                and my_dump_ext_opname not in self.my_dump.ext_opname_dataset_map.keys():
            failed_info.append([max(self.my_dump.get_all_orders()) + 1, 'NaN', golden_dataset, message])
            return False
        return True

    def _reverse_match_process(self: any, my_dump_ext_opname: str, golden_ext_opname: str,
                               golden_dataset: str, failed_info: list) -> bool:
        my_dump_dataset_path = self._construct_dataset_path(
            golden_ext_opname, golden_dataset, self.my_dump, my_dump_ext_opname)
        if not my_dump_dataset_path or not self.my_dump.have_dataset(my_dump_ext_opname, my_dump_dataset_path):
            order = self.my_dump.get_order_by_ext_opname(my_dump_ext_opname)
            op_name = self.get_original_opname(my_dump_ext_opname)
            message = 'No data match with {} in my dump data.'.format(golden_dataset)
            failed_info.append([order, op_name, golden_dataset, message])
            return False
        return True
